class ParamCardWriter(object):

    header = \
    """######################################################################\n""" + \
    """## PARAM_CARD AUTOMATICALY GENERATED BY THE UFO  #####################\n""" + \
    """######################################################################\n"""

    def __init__(self, filename, list_of_parameters=None, generic=False):
        """write a valid param_card.dat"""

        if not list_of_parameters:
            from parameters import all_parameters
            list_of_parameters = [param for param in all_parameters if param.nature == 'external']

        self.generic_output = generic
        if generic:
            self.define_not_dep_param(list_of_parameters)

        self.fsock = open(filename, 'w')
        self.fsock.write(self.header)

        self.write_card(list_of_parameters)

    def define_not_dep_param(self, list_of_parameters):
        """define self.dep_mass and self.dep_width in case that they are requested in the param_card.dat"""
        from particles import all_particles

        self.dep_mass = [(part, part.mass) for part in all_particles
                if part.pdg_code > 0 and part.mass not in list_of_parameters]
        self.dep_width = [(part, part.width) for part in all_particles
                if part.pdg_code > 0 and part.width not in list_of_parameters]

    @staticmethod
    def order_param(obj1, obj2):
        """ order parameter of a given block """

        maxlen = min([len(obj1.lhacode), len(obj2.lhacode)])

        for i in range(maxlen):
            if obj1.lhacode[i] < obj2.lhacode[i]:
                return -1
            elif obj1.lhacode[i] == obj2.lhacode[i]:
                return 0
            else:
                return 1
        # identical up to the first finish
        if len(obj1.lhacode) > len(obj2.lhacode):
            return 1
        elif len(obj1.lhacode) == len(obj2.lhacode):
            return 0
        else:
            return -1

    def write_card(self, all_ext_param):
        # list all lhablock
        all_lhablock = set([param.lhablock for param in all_ext_param])

        # ordonate lhablock alphabeticaly
        all_lhablock = list(all_lhablock)
        all_lhablock.sort()
        # put at the beginning SMINPUT + MASS + DECAY
        for name in ['DECAY', 'MASS','SMINPUTS']:
            if name in all_lhablock:
                all_lhablock.remove(name)
                all_lhablock.insert(0, name)

        for lhablock in all_lhablock:
            self.write_block(lhablock)
            need_writing = [ param for param in all_ext_param if
                    param.lhablock == lhablock]
            from functools import cmp_to_key
            need_writing.sort(key=cmp_to_key(self.order_param))
            [self.write_param(param, lhablock) for param in need_writing]

            if self.generic_output:
                if lhablock in ['MASS', 'DECAY']:
                    self.write_dep_param_block(lhablock)

        if self.generic_output:
            self.write_qnumber()

    def write_block(self, name):
        """ write a comment for a block"""

        self.fsock.writelines(f"""
        ###################################
        ## INFORMATION FOR {name.upper}
        ###################################
        """)
        if name != 'DECAY':
            self.fsock.write("""Block %s \n""" % name)

    def write_param(self, param, lhablock):

        lhacode = ' '.join(['%3s' % key for key in param.lhacode])
        if lhablock != 'DECAY':
            text = """  %s %e # %s \n""" % (lhacode, complex(param.value).real, param.name )
        else:
            text = '''DECAY %s %e \n''' % (lhacode, complex(param.value).real)
        self.fsock.write(text)

    def write_dep_param_block(self, lhablock):
        import cmath  # noqa: F401
        from parameters import all_parameters
        for parameter in all_parameters:
            try:
                exec("%s = %s" % (parameter.name, parameter.value))
            except Exception:
                pass
        text = "##  Not dependent paramater.\n"
        text += "## Those values should be edited following analytical the \n"
        text += "## analytical expression. Some generator could simply ignore \n"
        text += "## those values and use the analytical expression\n"

        if lhablock == 'MASS':
            data = self.dep_mass
            prefix = " "
        else:
            data = self.dep_width
            prefix = "DECAY "
        for part, param in data:
            if isinstance(param.value, str):
                value = complex(eval(param.value)).real
            else:
                value = param.value

            text += """%s %s %f # %s : %s \n""" % (prefix, part.pdg_code,
                        value, part.name, param.value)
        self.fsock.write(text)

    sm_pdg = [1,2,3,4,5,6,11,12,13,13,14,15,16,21,22,23,24,25]
    data = """Block QNUMBERS %(pdg)d  # %(name)s
        1 %(charge)d  # 3 times electric charge
        2 %(spin)d  # number of spin states (2S+1)
        3 %(color)d  # colour rep (1: singlet, 3: triplet, 8: octet)
        4 %(antipart)d  # Particle/Antiparticle distinction (0=own anti)\n"""

    def write_qnumber(self):
        """ write qnumber """
        from particles import all_particles

        text = """#===========================================================\n"""
        text += """# QUANTUM NUMBERS OF NEW STATE(S) (NON SM PDG CODE)\n"""
        text += """#===========================================================\n\n"""

        for part in all_particles:
            if part.pdg_code in self.sm_pdg or part.pdg_code < 0:
                continue
            text += self.data % {'pdg': part.pdg_code,
                                 'name': part.name,
                                 'charge': 3 * part.charge,
                                 'spin': 2 * part.spin + 1,
                                 'color': part.color,
                                 'antipart': part.name != part.antiname and 1 or 0}

        self.fsock.write(text)

# if '__main__' == __name__:
#     ParamCardWriter('./param_card.dat', generic=True)
#     print('write ./param_card.dat')



#!/bin/python
from __future__ import division
from anyBSM.utils import fieldtypes
from anyBSM.loopfunctions import A0, B0, B00, dB0, dB00, B1, dB1, C0, C1, C2, eps # noqa: F401
from anyBSM.ufo.function_library import complexconjugate, I # noqa: F401
import logging
import functools

__doc__ = """ # AnyDiagram
A module to store the analytic results of Feynman diagrams populated
with fields that couple with the most general Lorentz structure
without assuming any relations between the field masses and
couplings.\n

The analytic results are expressed in terms of the generic couplings,
masses and Passarino Veltman functions.
"""

logger = logging.getLogger('anyBSM.diagrams')

def multIply(coup, fac = 1):
    """ Multiply Feynman rule (UFO coupling) with -I to receive the
    parameter from the Lagrangian """
    try:
        return -I*fac*coup # numerical mode
    except:
        return f'-I*({fac})*({coup})' # analytical mode

def getCouplingNum(coupling, attribute = 'nvalue'):
    """ Get the (numerical) coupling from the `anyBSM.getcoupling()`-result"""
    coup = {i : multIply(c.get(attribute), coupling.get('sign', 1)) for i,c in coupling.items() if i != 'sign'}
    return coup.get('c',coup)

def iszero(CLR):
    """ check if a vertex (i.e. all its couplings) `CLR` is numerically negligible """
    if isinstance(CLR, dict):
        return all([abs(c) < eps for c in CLR.values()])
    return abs(CLR) < eps

def conj(CLR):
    """ Take the complex conjugate of a vertex (in case of a
    non-scalar vertex, all couplings are conjugated) """
    if isinstance(CLR, dict):
        return {k: conj(v) for k,v in CLR.items()}
    if isinstance(CLR, str):
        return f'complexconjugate({CLR})'
    return CLR.conjugate()

def diagram(Next=0,Nint=0,Nvert=0):
    """ Decorator for the actual generic diagrams. It ensures that the
    diagram with `Next`/`Nint` external/internal fields and `Nvert`
    vertices receives the correct inputs (number of couplings, masses
    etc) and prepares all information ready to be used in the generic
    analytic expressions.

    @diagram passes a dictionary of the form:
    ```python
    d = {
        'external' : [ externalfield1, ... ],
        'momenta'  : [ p12, p22, ...],
        'couplings' : [ CSSS, {'L': CFFSL, 'R': CFFSR}, ... ],
        'conj_couplings' : [ CS*S*S*, {'L': CFbFbS*L, 'R': CFbFbS*R}, ... ],
        'internal' : [ internalfield1, ... ],
        'masses' : [ MassOfinternalfield1, ... ],
        'factor' : symmetryfactor*colorfactor*loopfactor
    }
    ```
    to the actual generic diagram.
    """
    def wrapper2(diag):
        @functools.wraps(diag, updated=())
        def wrapper(self,**kwargs):
            all_args = ['external', 'factor', 'coupling_objects', 'conj_coupling_objects', 'internal', 'momenta']
            for arg in all_args:
                assert arg in kwargs , f"{diag.__name__}: missing argument {arg} in {kwargs})"
            assert Next == len(kwargs['external']), f"{diag.__name__}"
            assert Nint == len(kwargs['internal']), f'{diag.__name__}'
            assert Nvert == len(kwargs['coupling_objects']), f'{diag.__name__}'
            assert Nvert == len(kwargs['conj_coupling_objects']), f'{diag.__name__}'
            couplings_num = [getCouplingNum(c, self.coupling_values) for c in kwargs['coupling_objects']]
            conj_couplings_num = [getCouplingNum(c, self.coupling_values) for c in kwargs['conj_coupling_objects']]
            if self.anyzero(*couplings_num) or self.anyzero(*conj_couplings_num):
                return self.zero
            inttype = fieldtypes(kwargs['internal'])
            exttype = fieldtypes(kwargs['external'])
            masses = [p.get(self.mass_values) for p in kwargs['internal']]
            if 'derivative' not in kwargs:
                kwargs['derivative'] = False
            result = diag(self,**kwargs, couplings = couplings_num, conj_couplings = conj_couplings_num, inttype = inttype, exttype = exttype, masses = masses)
            if result is None:
                logger.error(f"{diag.__name__} with external={exttype} and internal={inttype} not (correctly) implemented!")
                return self.zero
            if not result:
                result = self.zero
            else:
                if self.evaluation == 'numerical' and isinstance(result, str):
                    result  = eval(result)
                elif not (result.startswith('+') or result.startswith('-')):
                    result = '+' + result
            logger.debug(f"{diag.__name__}({list(kwargs['internal'])}) = {result}")
            return result
        return wrapper
    return wrapper2

class GenericDiagrams():
    """
        Defines results for all generic amplitudes.
        Each class method with the @`diagram()` decorator is supposed
        to return a string with the masses and couplings inserted into
        the generic diagram it represents.\n
        Example:
        ```python
        GenericDiagrams.OnePointA(
            momenta = [],
            external = [h1],
            internal = [u3],
            coupling_objects = [{"L": GC_79, "R": GC_80}],
            derivative = False,
            factor = '1/(16*pi**2)')
        ```
        calculates the top quark contribution to the Higgs boson
        tadpole.
    """

    def __init__(self, evaluation = 'abbreviations', MSDR = 1):
        """
            * `evaluation`: evaluation mode as in anyBSM.anyBSM
            * `MSDR`: whether to work in $\\overline{\\text{MS}}$ (1) or
            $\\overline{\\text{DR}}$ (0).
        """
        self.evaluation = evaluation
        evaluation_methods = {
                'numerical': { 'couplings': 'nvalue', 'masses': 'nmass' },
                'analytical': { 'couplings': 'value', 'masses': 'mass' },
                'abbreviations': { 'couplings': 'name', 'masses': 'mass' }
                }
        self.coupling_values = evaluation_methods[evaluation]['couplings']
        self.mass_values = evaluation_methods[evaluation]['masses']
        self.zero = 0 if self.evaluation == 'numerical' else ''
        self.MSDR = MSDR

    def anyzero(self, *coups):
        """ Check if any of the vertices is numerically negligible
        i.e. whether the diagram is (numerically) zero """
        if self.evaluation in ['analytical', 'abbreviations']:
            return False
        return any([iszero(c) for c in coups])

    # 1L tadpoles
    @diagram(Next=1,Nint=1,Nvert=1)
    def OnePointA(self, **d):
        """ One-loop tadpole diagram
            * zero for everything but scalars
            * in the loop: $S=0,1/2,1$
        """
        if d['exttype'] != 'S':
            return
        fac = f"-1*({d['factor']})"
        m = d['masses'][0]
        c = d['couplings'][0]
        if d['inttype'] == 'F':
            return f"-1*{fac}*(({c['L']} + {c['R']})*({m})*A0({m}**2))"
        if d['inttype'] == 'V':
            return f"{fac}*({c})*({self.MSDR}*{m}**2-2*A0({m}**2))"
        if d['inttype'] in ['S', 'U']:
            fac = f'{fac}/2'
            fac = f'-1*{fac}' if d['inttype'] == 'U' else fac
            return f'+{fac}*({c})*A0({m}**2)'

    # 1L selfenergies
    @diagram(Next=2,Nint=1,Nvert=1)
    def TwoPointA(self, **d):
        """ One-loop selfenergy diagram with one four-point
        interaction.\n
         * internal: $S=0,1$
         * external: $S=0,1$
        """
        c = d['couplings'][0]
        m = d['masses'][0]
        fac = f"1/2*({d['factor']})"
        if d['derivative']:
            return self.zero

        if d['exttype'] == 'SS': # scalar selfenergy
            if d['inttype'] == 'S':
                return f'+({fac})*({c})*A0({m}**2)'
            if d['inttype'] == 'V':
                return f'-({fac})*({c})*4*(A0({m}**2)-{self.MSDR}*{m}**2/2)'

        if d['exttype'] == 'VV': # transverse part of vector boson selfenergy
            if d['inttype'] == 'S':
                return f'+({fac})*({c})*A0({m}**2)'
            if d['inttype'] == 'V':
                return f"-({fac})*((4*{c['1']}+{c['2']}+{c['3']})*A0({m}**2) - 2*{self.MSDR}*{m}**2*{c['1']})"

    @diagram(Next=2,Nint=2,Nvert=2)
    def TwoPointB(self, **d):
        """ One-loop selfenergy diagram with two three-point
        interactions as well as its derivative w.r.t the external
        momentum.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: $S=0,1/2,(1$ not yet checked$)$"""
        m1,m2 = d['masses']
        c1,c2 = d['couplings']
        c2 = conj(c2)
        if d['inttype'] == 'FF' or d['exttype'] == 'FF':
            c1L, c1R, c2L, c2R = c1['L'], c1['R'], c2['L'], c2['R']
        if d['inttype'] == 'UU':
            if d['exttype'] == 'VV':
                c1UUV1, c1UUV2, c2UUV1, c2UUV2 = c1['UUV1'], c1['UUV2'], c2['UUV1'], c2['UUV2']
            if d['exttype'] == 'SV':
                c2UUV1, c2UUV2 = c2['UUV1'], c2['UUV2']
            if d['exttype'] == 'VS':
                c1UUV1, c1UUV2 = c1['UUV1'], c1['UUV2']
        s = d['momenta'][0]
        fac = f"1/2*({d['factor']})"
        if d['exttype'] == 'SS': # scalar selfenergy
            fac = f"4*{fac}" if d['inttype'] == 'VV' else fac
            fac = f"-1*{fac}" if d['inttype'] == 'UU' else fac

            if d['derivative']:  # derivative w.r.t s
                # if d['inttype'] in ['SV', 'VS']:
                if d['inttype'] == 'SV':
                    dF0 = f'-2*B0({s},{m1}**2,{m2}**2)-(2*{s} + 2*{m1}**2-{m2}**2)*dB0({s},{m1}**2,{m2}**2)'
                    return f'-2*({fac})*({c1})*({c2})*({dF0})'
                if d['inttype'] == 'VS':
                    return self.zero
                if d['inttype'] == 'FF':
                    dG0 = f'({s}-{m1}**2-{m2}**2)*dB0({s},{m1}**2,{m2}**2) + B0({s},{m1}**2,{m2}**2)'
                    return f'-({fac})*(({c1L}*{c2L} + {c1R}*{c2R})*({dG0})-2*({c1L}*{c2R} + {c1R}*{c2L})*({m1})*({m2})*dB0({s},{m1}**2,{m2}**2))'
                if d['inttype'] in ['SS', 'VV', 'UU']:
                    if d['inttype'] == 'UU':
                        c2 = conj(c2)
                    return f'-{fac}*({c1})*({c2})*dB0({s},{m1}**2,{m2}**2)'

            # if d['inttype'] in ['SV', 'VS']:
            if d['inttype'] == 'SV':
                F0 = f'A0({m1}**2)-2*A0({m2}**2)-(2*{s} + 2*{m1}**2-{m2}**2)*B0({s},{m1}**2,{m2}**2)'
                return f'-2*({fac})*({c1})*({c2})*({F0})'
            if d['inttype'] == 'VS':  # set to 0 to avoid double counting
                # F0 = f'A0({m2}**2)-2*A0({m1}**2)-(2*{s} + 2*{m2}**2-{m1}**2)*B0({s},{m2}**2,{m1}**2)'
                return self.zero # f'+2*{fac}*({c1})*({c2})*({F0})'
            if d['inttype'] == 'VV':
                return f'-{fac}*({c1})*({c2})*(B0({s},{m1}**2,{m2}**2)-{self.MSDR}/2)'
            if d['inttype'] in ['SS', 'UU']:
                if d['inttype'] == 'UU':
                    c2 = conj(c2)
                return f'-{fac}*({c1})*({c2})*B0({s},{m1}**2,{m2}**2)'
            if d['inttype'] == 'FF':
                G0 = f'(({s}-{m1}**2-{m2}**2)*B0({s},{m1}**2,{m2}**2) - A0({m1}**2) - A0({m2}**2))'
                return f'-({fac})*((({c1L})*({c2L}) + ({c1R})*({c2R}))*({G0}) - 2*(({c1L})*({c2R}) + ({c1R})*({c2L}))*({m1})*({m2})*B0(({s}),({m1})**2,({m2})**2))'

        if d['exttype'] == 'VV': # transverse part of vector selfenergy
            if d['derivative']: # derivative w.r.t. s
                dB0012 = f'dB00({s},{m1}**2,{m2}**2)'
                dB012 = f'dB0({s},{m1}**2,{m2}**2)'
                if d['inttype'] == 'SS':
                    return f'-4*{fac}*({c1})*({c2})*{dB0012}'
                if d['inttype'] == 'VV':
                    return f'-1*{fac}*({c1})*({c2})*(10*({dB0012}) + ({m2}**2 + {m1}**2 + 4*{s})*({dB012}) + 4*B0({s},{m1}**2,{m2}**2)+2/3*{self.MSDR})'
                if d['inttype'] == 'UU':
                    return f'+({fac})*(({c1UUV1})-({c1UUV2}))*(({c2UUV1})-({c2UUV2}))*{dB0012}'
                # if d['inttype'] == 'SV':
                if d['inttype'] in ['SV', 'VS']:
                    return f'+({fac})*({c1})*({c2})*{dB012}'
                # 16.3.2023: revert to having both SV and VS cases (factor 2 removed above)
                # if d['inttype'] == 'VS':
                #    return self.zero
                if d['inttype'] == 'FF':
                    db0 = f'dB0({s},{m1}**2,{m2}**2)'
                    b0 = f'B0({s},{m1}**2,{m2}**2)'
                    dH0 = f'4*dB00({s},{m1}**2,{m2}**2) + ({s} - {m1}**2 - {m2}**2)*{db0} + {b0}'
                    return f'{fac}*((({c1L})*({c2L}) + ({c1R})*({c2R}))*({dH0}) + 2*(({c1L})*({c2R}) + ({c1R})*({c2L}))*({m1})*({m2})*{db0})'

            B0012 = f'B00({s},{m1}**2,{m2}**2)'
            B012 = f'B0({s},{m1}**2,{m2}**2)'
            if d['inttype'] == 'SS':
                return f'-4*{fac}*({c1})*({c2})*{B0012}'
            if d['inttype'] == 'VV':
                return f'-1*{fac}*({c1})*({c2})*(10*{B0012} + ({m2}**2 + {m1}**2 + 4*{s})*{B012} + A0({m1}**2) + A0({m2}**2) - 2*{self.MSDR}*({m1}**2+{m2}**2-{s}/3))'
            if d['inttype'] == 'UU':
                return f'{fac}*(({c1UUV1})-({c1UUV2}))*(({c2UUV1})-({c2UUV2}))*{B0012}'
            if d['inttype'] in ['SV', 'VS']:
                return f'+({fac})*({c1})*({c2})*{B012}'
            # if d['inttype'] == 'VS':
            #     return self.zero
            if d['inttype'] == 'FF':
                b0 = f'B0({s},{m1}**2,{m2}**2)'
                H0 = f'4*B00({s},{m1}**2,{m2}**2)-A0({m1}**2)-A0({m2}**2) + ({s} - {m1}**2 - {m2}**2)*{b0}'
                return f'{fac}*((({c1L})*({c2L}) + ({c1R})*({c2R}))*({H0}) + 2*(({c1L})*({c2R}) + ({c1R})*({c2L}))*({m1})*({m2})*({b0}))'

        if d['exttype'] == 'SV' or 'VS': # scalar--gauge-boson mixing self-energies
            if d['derivative']:
                return
                # not implemented (not needed?)
            B012 = f'B0({s},{m1}**2,{m2}**2)'
            B112 = f'B1({s},{m1}**2,{m2}**2)'
            B121 = f'B1({s},{m2}**2,{m1}**2)'
            if d['inttype'] == 'SS':
                return f'+({fac})*({c1})*({c2})*(({B012})+2*({B112}))'
            if d['inttype'] == 'FF':
                return (
                        f'+2*({fac})*('
                        f'(({c1L})*({c2L}) +({c1R})*({c2R})) *({m1})*({B012})'
                        f'+(({c1L})*({c2L})*({m1}) +({c1R})*({c2R})*({m1}) +({c1R})*({c2L})*({m2}) +({c1L})*({c2R})*({m2}))*({B112})'
                        f')'
                       )
            if d['inttype'] == 'VV':
                return f'-3*({fac})*({c1})*({c2})*(({B012})+2*({B112}))'
            if d['inttype'] == 'SV':
                return f'-({fac})*({c1})*({c2})*(({B012})-({B112}))'
            if d['inttype'] == 'VS':
                return f'-({fac})*({c1})*({c2})*(({B012})-({B121}))'
            #    return self.zero # put separately from 'SV' case?
            if d['inttype'] == 'UU' and d['exttype'] == 'SV':
                # c1 is SUU coupling
                # c2 is UUV coupling which comes as dict with keys 'UUV1', 'UUV2', 'sign'
                return f'+({fac})*({c1})*(({c2UUV1})*({B012})+(({c2UUV1})-({c2UUV2}))*({B112}))' # not checked yet
            if d['inttype'] == 'UU' and d['exttype'] == 'VS':
                # c2 is SUU coupling
                # c1 is UUV coupling which comes as dict with keys 'UUV1', 'UUV2', 'sign'
                return f'+({fac})*({c2})*(-({c1UUV2})*({B012})+(({c1UUV1})-({c1UUV2}))*({B112}))' # not checked yet

        if d['exttype'] == 'FF': # fermion selfenergy
            if d['inttype'] not in ['FV', 'FS']:
                return 0
            if d['external'][0] != d['external'][1].anti():
                # external states are in the mass basis -> no mixing
                mfext = 0
            else:
                mfext = d['external'][0].get(self.mass_values)
            if d['derivative']:
                b0 = f'dB0({s},{m1}**2,{m2}**2)'
                b1 = f'dB1({s},{m1}**2,{m2}**2)'
            else:
                if d['inttype'] == 'FV':
                    b0 = f'(B0({s},{m1}**2,{m2}**2) - 1/2*{self.MSDR})'
                    b1 = f'(B1({s},{m1}**2,{m2}**2) + 1/2*{self.MSDR})'
                elif d['inttype'] == 'FS':
                    b0 = f'B0({s},{m1}**2,{m2}**2)'
                    b1 = f'B1({s},{m1}**2,{m2}**2)'

            SigSL = f'2*{m1}*({c1R})*({c2L})*{b0}'
            SigSR = f'2*{m1}*({c1L})*({c2R})*{b0}'
            SigVL = f'-2*({c1L})*({c2L})*{b1}'
            SigVR = f'-2*({c1R})*({c2R})*{b1}'
            if d['inttype'] == 'FV':
                SigSL = f'-4*({SigSL})'
                SigSR = f'-4*({SigSR})'
                SigVL = f'2*({SigVL})'
                SigVR = f'2*({SigVR})'
            return f'+({fac})*(({SigSL}+{SigSR})/2 + ({mfext})*({SigVL} + {SigVR})/2)'

        return

    @diagram(Next=2,Nint=2,Nvert=2)
    def TwoPointTA(self, **d):
        """ One-loop tadpole attached to a tree-level propagator, i.e.
        tadpole correction to a selfenergy.\n
         * internal: $S=0,1/2,1$
         * external: $S=0,1$ (1/2 not yet implemented)
        """
        m1,m2 = d['masses']
        c1,c2 = d['couplings']
        fac = f"1/2*({d['factor']})"

        if d['derivative'] or not d['inttype'].startswith('S'):
            return self.zero

        if d['exttype'] == 'SS': # scalar self-energy with tadpole insertion
            if d['inttype'] == 'SS':
                return f"+({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"
            if d['inttype'] == 'SF':
                c2L,c2R = c2['L'],c2['R']
                return f"-2*({fac})*({c1})*(({c2L})+({c2R}))*{m2}*A0({m2}**2)/({m1}**2)"
            if d['inttype'] == 'SV':
                return f"+2*({fac})*({c1})*({c2})*(({self.MSDR})*({m2}**2)-2*A0({m2}**2))/({m1}**2)"
            if d['inttype'] == 'SU':
                return f"-({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"

        if d['exttype'] == 'VV': # vector self-energy with tadpole insertion
            if d['inttype'] == 'SS':
                return f"+({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"
            if d['inttype'] == 'SF':
                c2L,c2R = c2['L'],c2['R']
                return f"-2*({fac})*({c1})*(({c2L})+({c2R}))*{m2}*A0({m2}**2)/({m1}**2)"
            if d['inttype'] == 'SV':
                return f"+2*({fac})*({c1})*({c2})*(({self.MSDR})*({m2}**2)-2*A0({m2}**2))/({m1}**2)"
            if d['inttype'] == 'SU':
                return f"-({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"

        if d['exttype'] == 'FF': # fermion selfenergy with tadpole insertion
            mS,mX = d['masses']
            cL, cR = d['couplings'][0]['L'], d['couplings'][0]['R']

            Stad = self.OnePointA(
                external = [d['internal'][0]],
                internal = [d['internal'][1]],
                coupling_objects = [d['coupling_objects'][1]],
                conj_coupling_objects = [d['conj_coupling_objects'][1]],
                factor = d['factor'],
                momenta = []
                )

            return f'1/({mS})**2*({Stad})*({cL} + {cR})/2'

        if d['exttype'] in ['SV','VS']:
            return

    # Three point functions
    @diagram(Next=3,Nint=0,Nvert=1)
    def ThreePointTree(self, **d):
        """ Tree-level S->SS diagram """
        try:
            return f"{d['couplings'][0]}"
        except IndexError:
            return ''

    @diagram(Next=3,Nint=2,Nvert=2)
    def ThreePointB(self, **d):
        """ One-loop correction to three-point functions with one
        quartic and one trilinear coupling.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: only $S=1/2$"""
        if d['exttype'] != 'SSS':
            return
        fac = d['factor']
        c1,c2 = d['couplings']
        m1,m2 = d['masses']
        s = d['momenta'][0]
        if d['inttype'] == 'SS':
            return f'+({fac})*({c1})*({c2})*B0({s},{m1}**2,{m2}**2)/2'
        if d['inttype'] == 'VV':
            return f'+({fac})*({c1})*({c2})*(2*B0({s},{m1}**2,{m2}**2)-{self.MSDR})'

    @diagram(Next=3,Nint=3,Nvert=3)
    def ThreePointC(self, **d):
        """ One-loop correction to three-point functions with three
        trilinear couplings.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: only $S=0$"""

        if d['exttype'] != 'SSS':
            return
        m1,m2,m3 = d['masses']
        fac = d['factor']
        c1, c2, c3 = d['couplings']
        p12,p22,p32 = d['momenta']
        if d['inttype'] in ['SSS', 'UUU', 'VVV']:
            fac = f'-1*{fac}' if d['inttype'] == 'UUU' else fac
            fac = f'-4*{fac}' if d['inttype'] == 'VVV' else fac
            return f'-1*{fac}*({c1})*({c2})*({c3})*C0({p12},{p32},{p22},{m1}**2,{m2}**2,{m3}**2)'
        if d['inttype'] == 'FFF':
            c1L,c1R = c1['L'],c1['R']
            c2L,c2R = c2['L'],c2['R']
            c3L,c3R = c3['L'],c3['R']
            return (
                    f'+({fac})* ('
                    f'2*B0({p32},{m2}**2, {m3}**2)* ({c1L}* ({c2L}* {c3R}* {m1} + {c2R}*{c3L}*{m2} + {c2R}*{c3R}*{m3})+ {c1R}*({c2R}*{c3L}*{m1} + {c2L}*{c3R}*{m2} + {c2L}*{c3L}*{m3}))'
                    f'+ {m1}*C0({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2)* (({c1R}*{c2R}*{c3L} + {c1L}*{c2L}*{c3R})* ({p12} + {p22} - {p32}) +2*{m2}*{m3}*({c1L}*{c2L}*{c3L} + {c1R}*{c2R}*{c3R}) +2*{m1}*({c1R}* ({c2R}*{c3L}*{m1} + {c2L}*{c3R}*{m2} + {c2L}*{c3L}*{m3}) + {c1L}*({c2L}*{c3R}*{m1} + {c2R}*{c3L}*{m2} + {c2R}*{c3R}*{m3})))'
                    f'+C1({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2)*(({p12}+{p22}-{p32})*(({c1R}*{c2R}*{c3L} + {c1L}*{c2L}*{c3R})*{m1} + ({c1L}*{c2R}*{c3L} + {c1R}*{c2L}*{c3R})*{m2}) +2*{p12}*({c1R}*{c3L}*({c2R}*{m1}+ {c2L}*{m3}) + {c1L}*{c3R}*({c2L}*{m1} + {c2R}*{m3})))'
                    f'+C2({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2)*(2*{p22}*({c1R}*{c2R}*{c3L}*{m1} + {c1L}*{c2L}*{c3R}*{m1} + {c1L}*{c2R}*{c3L}*{m2} + {c1R}*{c2L}*{c3R}*{m2})+ ({p12}+{p22}-{p32})*({c1R}*{c3L}*({c2R}*{m1} + {c2L}*{m3}) + {c1L}*{c3R}*({c2L}*{m1} + {c2R}*{m3})))'
                    f')'
                    )
        if d['inttype'] == 'SSV':
            return (
                    f'-1*({fac})*({c1})*({c2})*({c3})*(B0({p32}, {m2}**2, {m3}**2) - (({p12})-({p32})-({m1})**2)*C0({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2) + ({p12}-{p22}+{p32})*C1({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2) + ({p12}-{p22}-{p32})*C2({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2))'
                    )
        if d['inttype'] == 'SVS':
            return (
                    f'-1*({fac})*({c1})*({c2})*({c3})*(B0({p32}, {m2}**2, {m3}**2) - ({p22}-{p32}-{m1}**2)*C0({p22}, {p32}, {p12}, {m1}**2, {m3}**2, {m2}**2) - ({p12}-{p22}-{p32})*C1({p22}, {p32}, {p12}, {m1}**2, {m3}**2, {m2}**2) - ({p12}-{p22}+{p32})*C2({p22}, {p32}, {p12}, {m1}**2, {m3}**2, {m2}**2))'
                    )
        if d['inttype'] == 'VSS':
            return (
                    f'-1*({fac})*({c1})*({c2})*({c3})*(B0({p32}, {m2}**2, {m3}**2) -({p12}-{p22}+{p32}-{m1}**2)*C0({p32}, {p22}, {p12}, {m2}**2, {m3}**2, {m1}**2) - 2*({p12} -{p22})*C1({p32}, {p22}, {p12}, {m2}**2, {m3}**2, {m1}**2) -(3*{p12}+{p22}-{p32})*C2({p32}, {p22}, {p12}, {m2}**2, {m3}**2, {m1}**2))'
                    )
        if d['inttype'] == 'SVV':
            return (
                    f'+({fac})*({c1})*({c2})*({c3})*(2*B0({p32}, {m2}**2, {m3}**2) + ({p12}+{p22}-{p32}+2*{m1}**2)*C0({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2) - (3*{p12}+{p22}-{p32})*C1({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2) - ({p12}+3*{p22}-{p32})*C2({p12}, {p32}, {p22}, {m1}**2, {m2}**2, {m3}**2))/2'
                    )
        if d['inttype'] == 'VSV':
            return (
                    f'+({fac})*({c1})*({c2})*({c3})*(2*B0({p32}, {m2}**2, {m3}**2) - ({p12}+{p22}-{p32}-2*{m1}**2)*C0({p12}, {p22}, {p32}, {m2}**2, {m1}**2, {m3}**2) - (7*{p12}-{p22}+{p32})*C1({p12}, {p22}, {p32}, {m2}**2, {m1}**2, {m3}**2) - (3*{p12}-3*{p22}+5*{p32})*C2({p12}, {p22}, {p32}, {m2}**2, {m1}**2, {m3}**2))/2'
                    )
        if d['inttype'] == 'VVS':
            return (
                    f'+({fac})*({c1})*({c2})*({c3})*(2*B0({p32}, {m2}**2, {m3}**2) - ({p12}+{p22}-{p32}-2*{m1}**2)*C0({p22}, {p12}, {p32}, {m3}**2, {m1}**2, {m2}**2) + ({p12}-7*{p22}-{p32})*C1({p22}, {p12}, {p32}, {m3}**2, {m1}**2, {m2}**2) + (3*{p12}-3*{p22}-5*{p32})*C2({p22}, {p12}, {p32}, {m3}**2, {m1}**2, {m2}**2))/2'
                    )

    @diagram(Next=3,Nint=2,Nvert=2)
    def ThreePointTA(self, **d):
        """ One-loop tadpole attached to one of the external legs of a
        tree-level four point interaction.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: only $S=0$"""

        if d['exttype'] != 'SSS':
            return

        m1, m2 = d['masses']
        c1, c2 = d['couplings']
        fac = f"1/2*({d['factor']})"

        if d['inttype'] == 'SS':
            return f"-({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"
        if d['inttype'] == 'SF':
            c2L, c2R = c2['L'], c2['R']
            return f"+2*({fac})*({c1})*(({c2L})+({c2R}))*({m2})*A0({m2}**2)/({m1}**2)"
        if d['inttype'] == 'SV':
            return f"-2*({fac})*({c1})*({c2})*(({self.MSDR})*({m2}**2)-2*A0({m2}**2))/({m1}**2)"
        if d['inttype'] == 'SU':
            return f"+({fac})*({c1})*({c2})*A0({m2}**2)/({m1}**2)"

    @diagram(Next=3,Nint=3,Nvert=3)
    def ThreePointWFRT(self, **d):
        """ External leg correction to the three-point function with a
        tadpole insertion.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: only $S=0$"""
        if d['internal'][0].name == d['external'][-1].name: # diagonal WFR
            return self.zero
        SSs_coupling = d['couplings'][0]
        sS_selfenergyTA = self.TwoPointTA(
                external = [d['internal'][0], d['external'][-1]],
                internal = [d['internal'][1], d['internal'][2]],
                coupling_objects = [d['coupling_objects'][1], d['coupling_objects'][2]],
                conj_coupling_objects = [d['conj_coupling_objects'][1], d['conj_coupling_objects'][2]],
                factor = d['factor'],
                momenta = []
                )
        if not (SSs_coupling and sS_selfenergyTA):
            return self.zero
        mprop = d['masses'][0]
        return f'-1/({mprop}**2-({d["momenta"][2]}))*({SSs_coupling})*({sS_selfenergyTA})'

    @diagram(Next=3,Nint=2,Nvert=2)
    def ThreePointWFRA(self, **d):
        """ External leg correction to the three-point function with
        an A0-integral insertion.\n
         * internal: all possible combinations of $S=0,1$
         * external: only $S=0$"""
        if d['internal'][0].name in [d['external'][-1].name,d['external'][-1].anti().name]: # diagonal WFR
            return self.zero
        SSs_coupling = d['couplings'][0]
        sS_selfenergyA = self.TwoPointA(
                external = [d['internal'][0], d['external'][-1]],
                internal = [d['internal'][1]],
                coupling_objects = [d['coupling_objects'][1]],
                conj_coupling_objects = [d['conj_coupling_objects'][1]],
                factor = d['factor'],
                momenta = []
                )
        if not (SSs_coupling and sS_selfenergyA):
            return self.zero
        mprop = d['masses'][0]
        return f'-1/({mprop}**2-({d["momenta"][2]}))*({SSs_coupling})*({sS_selfenergyA})'

    @diagram(Next=3,Nint=3,Nvert=3)
    def ThreePointWFRB(self, **d):
        """ External leg correction to the three-point function with
        an B0-integral insertion.\n
         * internal: all possible combinations of $S=0,1/2,1$
         * external: only $S=0$"""

        diagonal = (d['internal'][0].name in [d['external'][-1].name,d['external'][-1].anti().name])
        SSs_coupling = d['couplings'][0]
        sS_selfenergyB = self.TwoPointB(
                external = [d['internal'][0], d['external'][-1]],
                internal = [d['internal'][1], d['internal'][2]],
                coupling_objects = [d['coupling_objects'][1], d['coupling_objects'][2]],
                conj_coupling_objects = [d['conj_coupling_objects'][1], d['conj_coupling_objects'][2]],
                factor = d['factor'],
                momenta = [d['momenta'][2], d['momenta'][2]],
                derivative = diagonal
                )
        if not (SSs_coupling and sS_selfenergyB):
            return self.zero
        if diagonal:
            return f'1/2*({SSs_coupling})*({sS_selfenergyB})'
        mprop = d['masses'][0]
        return f'-1/({mprop}**2-({d["momenta"][2]}))*({SSs_coupling})*({sS_selfenergyB})'



from os import path
import os
import yaml
import appdirs
import logging
from typing import Union
import git
from anyBSM.locks import safe_open

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

UPSTREAM = 'https://gitlab.com/anybsm/anybsm_models.git'

def get_config(ask: bool = True) -> dict:
    """Read configuration file and return its contents
    The directory containing the `anyBSM_config.yaml` file can be specified
    using the environment variable `anyBSM_CONF_DIR`. If it is not set,
    the location is determined automatically using `appdirs.user_config_dir('anyBSM')` i.e. most likely
    `~/.config/anyBSM` under Linux/Unix and
    ` ~/Library/Application Support/anyBSM` under Mac OS.
    """
    config_dir = os.environ.get('anyBSM_CONF_DIR', appdirs.user_config_dir('anyBSM'))
    config_file = path.join(config_dir, 'anyBSM_config.yaml')
    if not path.isfile(config_file):
        create_user_config(config_file, models_dir = None, ask = ask)
    with safe_open(config_file, 'r') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
        if 'models_dir' not in config:
            logger.error(f"Invaild config file: {config_file}. Variable 'models_dir' not found.")
        return config

def create_user_config(config_file: str, models_dir: Union[None, str] = None, ask: bool = False):
    """Create the user's config file and ask for path to model directory.

    Args:
        config_file: location of the config file `anyBSM_config.yaml`.
        models_dir: path to directory where models should be stored. If None, user is asked or default directory is chosen.
        ask: whether to query the user for a directory
    """
    # setup directory for model files
    config_dir = os.path.dirname(config_file)
    default_models_dir = path.join(config_dir, 'models')
    if models_dir is None:
        if ask:
            models_dir = input(f"Enter path to directory where model files should be stored (press enter for default: {default_models_dir}):") or default_models_dir
        else:
            models_dir = default_models_dir
    models_dir = os.path.expanduser(models_dir)
    if not path.exists(models_dir):
        os.makedirs(models_dir)
    else:
        if not path.isdir(models_dir):
            raise NotADirectoryError(models_dir)
    logger.info(f"storing model files at: {models_dir}")
    # download models if dir is empty
    if not os.listdir(models_dir):
        download_models(models_dir)
    # save config
    config = {}
    config['models_dir'] = models_dir
    os.makedirs(os.path.dirname(config_file), exist_ok=True)
    with safe_open(config_file, 'w') as f:
        yaml.dump(config, f)
    logger.info(f"storing config file at: {config_file}")

def download_models(models_dir: str):
    """ Clone model files from git repo

    Args:
        models_dir: download target path
    """
    if not path.isdir(models_dir):
        raise NotADirectoryError(models_dir)
    logger.info(f"Trying to download model files from {UPSTREAM}...")
    try:
        # are we installing from dev-branch in CI?
        branch = 'develop' if os.environ.get('CI_COMMIT_REF_NAME', 'main') != 'main' else 'main'
        git.Repo.clone_from(UPSTREAM, models_dir, branch = branch)
    except Exception as e:
        logger.warning(f'Failed to clone model files. Error was:\n{e}')
    logger.info(f"Trying to download model files from {UPSTREAM}... done.")



from __future__ import division
from functools import lru_cache
import cmath
import logging
import re

logger = logging.getLogger('anyBSM.loopfunctions')
# from functools import lru_cache
try:
    import pyCollier
    collier = True
except Exception as e:
    collier = False
    logger.error(str(e))
    logger.warning("Could not import collier. Evaluation with finite external momentum not possible!")

__doc__ = """ # AnyLoopFunction
A module to store all needed loop functions.\n
For $p^2=0$ we implement the results for the loop functions directly
(see below).
For finite external momentum we rely on
[COLLIER](https://collier.hepforge.org/index.html) [Denner, Dittmaier,
Hofer] via linked [pycollier](https://anybsm.gitlab.io/pycollier/pyCollier.html).
"""

global qren
qren = None
""" The renormalization scale $Q_{ren.}$ (not the squared one!) """

eps = 10**-5
""" When two masses are assumed to be equal """

epszero = 10**-8
""" When masses are assumed to vanish """

global loop_functions
loop_functions = []


def loopfunction():
    """ decorator for loop functions. Adds a loop function to the list `loop_functions` and enables caching """
    def decorator(func):
        func = lru_cache(maxsize=5000)(func)
        loop_functions.append(func)
        return func
    return decorator

def clear_loopfunction_cache():
    global loop_functions
    for func in loop_functions:
        func.cache_clear()
UVparts = [
        (              r"A0\((.*?)\)",                   r"\1"), #          A0(m)_div = m
        (              r"B0\((.*?)\)",                    r"1"), #   B0(p2,m1,m2)_div = 0
        (             r"dB0\((.*?)\)",                    r"0"), #  dB0(p2,m1,m2)_div = 1
        (              r"B1\((.*?)\)",               r"(-1/2)"), #   B1(p2,m1,m2)_div = - 1/2
        (             r"dB1\((.*?)\)",                    r"0"), #  dB1(p2,m1,m2)_div = 0
        ( r"B00\((.*?),(.*?),(.*?)\)",  r"((\2+\3)/4 - \1/12)"), #  B00(p2,m1,m2)_div = (m1 + m2)/4 - p2/12
        (            r"dB00\((.*?)\)",              r"(-1/12)"), # dB00(p2,m1,m2)_div = - 1/12
        (              r"C0\((.*?)\)",                    r"0"), #        C0(...)_div = 0
        (              r"C1\((.*?)\)",                    r"0"), #        C1(...)_div = 0
        (              r"C2\((.*?)\)",                    r"0"), #        C2(...)_div = 0
        (              r"D0\((.*?)\)",                    r"0"), #        D0(...)_div = 0
        ]
""" UV divergent parts of one-loop integrals (represented as regular expressions) """

UVfin = [ f[0].replace(r'\((.*?)\)','') for f in UVparts if f[1] == r"0" ]
""" list of one-loop integrals which are UV finite """

def get_UV_part(expr: str):
    """ Replaces all loop functions appearing in  `expr` by their UV-divergent part.

    Args:
        expr: expression to process.

    Returns:
        Expression with loop functions replaced by UV-divergent part.
    """
    for repl in UVparts:
        expr = re.sub(repl[0], repl[1], expr)
    return expr

def set_renscale(q):
    """ sets the renormalization scale in pycollier as well as the
    globally used variable `qren`. Note that pycollier's
    set_renscale-function takes the squared value while ours does not.
    (Re)setting the renormalization scale clears the cached loop functions."""
    global qren
    qren = q
    if collier:
        pyCollier.set_renscale(qren**2)
    for func in loop_functions:
        if func.__name__ in UVfin:
            continue
        func.cache_clear()

def set_delta(delta):
    """ sets the UV regulator in pycollier (delta = 0 by default).
    Changing this value also clears the cached loop functions."""
    if collier:
        pyCollier.set_delta(delta)
    clear_loopfunction_cache()

def collerr(p2):
    """ Throws error message if a loop integral with $p^2\\neq0$ was
    issued while COLLIER is not installed/found"""
    if abs(p2) > eps:
        raise ModuleNotFoundError('Evaluation with finite external momentum requires pyCollier!')

@loopfunction()
def lnbar(x):
    """ $\\overline{\\log}x = \\log \\frac{x}{Q_{ren.}^2}$ """
    return cmath.log(x/qren**2)

@loopfunction()
def A0(x):
    """ $A_0(x) = x(1-\\overline{\\log}x$"""
    if abs(x) < epszero:
        return 0
    return x*(1-lnbar(x))

def B0_p0(x,y):
    """ $B_0(p^2=0;x,y)=\\frac{A_0(x)-A_0(y)}{x-y}$ """
    if abs(x) < epszero and abs(y) < epszero:
        return 0
    if abs(x-y)/abs(x+y) < eps:
        return -lnbar(x)
    return (A0(x)-A0(y))/(x-y)

@loopfunction()
def B0(p2,x,y):
    """ $B_0(p^2;x,y)$ wrapper for pycollier"""
    if collier and abs(p2) > eps:
        return (pyCollier.b0(p2, x, y))
    collerr(p2)
    return B0_p0(x,y)

def B0p_p0(x,y):
    """ $\\frac{\\partial}{\\partial x}B_0(0;x,y) = -\\frac{1}{x-y}\\left(\\overline{\\log}x + B_0(0;x,y) \\right) $"""
    if abs(x) < epszero:
        if abs(y) < epszero:
            return 0
        logger.error('B0p_p0(0,y) is IR divergent!')
    if abs(y)/abs(x) < epszero:
        return -1/x
    if abs(x-y)/abs(x+y) < eps:
        return -1/(2*x)
    return -(lnbar(x)+B0_p0(x,y))/(x-y)

@loopfunction()
def B1(p2,x,y):
    """ $B_1(p^2,x,y)$ wrapper for pycollier """
    return (pyCollier.b1(p2, x, y))

@loopfunction()
def dB1(p2,x,y):
    """ $\\dot{B}_1(p^2,x,y)$ wrapper for pycollier """
    return (pyCollier.db1(p2, x, y))

@loopfunction()
def B00(p2,x,y):
    """ $B_{00}(p^2,x,y)$ wrapper for pycollier """
    if collier and abs(p2) > eps:
        return (pyCollier.b00(p2, x, y))
    collerr(p2)
    return B00_p0(x,y)

def B00_p0(x,y):
    """ $B_{00}(p^2=0,x,y) = \\frac{1}{4} \\left(x B_0(0;x,y) + A_0(y) + \\frac{x+y}{2}\\right)$ """
    return x*B0_p0(x,y)/4 + A0(y)/4 + (x + y)/8

def dB0_p0(x,y):
    """ $\\frac{\\partial}{\\partial p^2} B_0(p^2;x,y)|_{p^2=0} = \\frac{x^2 + 2 A_0(x) y - y^2 - 2 x A_0(y)}{2 (x-y)^3}$"""
    [x,y] = sorted([abs(x),abs(y)])
    if abs(y) < epszero:
        return 0
    if abs(x-y)/abs(x+y) < eps:
        return 1/(6*x)
    return (x**2+2*A0(x)*y-y**2-2*x*A0(y))/(2*(x-y)**3)

@loopfunction()
def dB0(p2,x,y):
    """ $\\dot{B}_0(p^2,x,y)$ wrapper for pycollier """
    if collier and abs(p2) > eps:
        return (pyCollier.db0(p2, x, y))
    collerr(p2)
    return dB0_p0(x,y)

def dB00_p0(x,y):
    """ $\\frac{\\partial}{\\partial p^2} B_{00}(p^2;x,y)|_{p^2=0} =\
        \\frac{-5x^3 + 27x^2y - 27xy^2 + 5y^3 +\
        6(x-y)^3 \\overline{\\log}y + \
        6x^2*(x-3y)\\log\\frac{x}{y}}{72(x-y)^3}\
       \\frac{x^2 + 2 A_0(x) y - y^2 - 2 x A_0(y)}{2 (x-y)^3}$"""
    [x,y] = sorted([abs(x),abs(y)])
    if abs(y) < epszero:
        logger.error('dB00_p0(0,0) is IR divergent!')
        return 0
    if abs(x) < epszero:
        return -(5-6*lnbar(y))/72
    if abs(x-y)/abs(x+y) < eps:
        return lnbar(x)/12
    return (-5*x**3 + 27*x**2*y - 27*x*y**2 + 5*y**3 + 6*(x-y)**3*lnbar(y) + 6*x**2*(x-3*y)*cmath.log(x/y))/(72*(x-y)**3)

@loopfunction()
def dB00(p2,x,y):
    """ $\\dot{B}_{00}(p^2;x,y)$ wrapper for pycollier """
    if collier and abs(p2) > eps:
        return (pyCollier.db00(p2, x, y))
    collerr(p2)
    return dB00_p0(x,y)


@loopfunction()
def C0_p0(x,y,z):
    """ $$C_{0}(p^2=0;x,y,z) = \\frac{B_0(0;y,x)-B_0(0;z,x)}{y-z}$$"""
    if abs(y) < epszero and abs(z) < epszero:
        if abs(x) < epszero:
            return 0
        logger.error('C0_p0(x,0,0) is IR divergent!')
        return 0

    if abs(abs(y)-abs(z))/abs(abs(y)+abs(z)) < eps:
        return B0p_p0(z,x)
    return (B0_p0(y,x)-B0_p0(z,x))/(y-z)

@loopfunction()
def C0(p12,p22,p2p3,x,y,z):
    """ $C_{0}(p^2;x,y,z)$ wrapper for pycollier """
    if collier and (abs(p12) > 0 or abs(p22) > 0 or abs(p2p3) > eps):
        return (pyCollier.c0(p12,p22,p2p3,x,y,z))
    collerr(p12+p22+p2p3)
    return C0_p0(x,y,z)

def C1_p0(x,y,z):
    """ $C_{1}(p^2=0;x,y,z)=0$ """
    return 0

@loopfunction()
def C1(p12,p22,p2p3,x,y,z):
    """ $C_{1}(p^2;x,y,z)$ wrapper for pycollier """
    if collier:
        return (pyCollier.c1(p12,p22,p2p3,x,y,z))
    collerr(p12+p22+p2p3)
    return C1_p0(x,y,z)

def C2_p0(x,y,z):
    """ $C_{2}(p^2=0;x,y,z)=0$ """
    return 0

@loopfunction()
def C2(p12,p22,p2p3,x,y,z):
    """ $C_{2}(p^2;x,y,z)$ wrapper for pycollier """
    if collier:
        return (pyCollier.c2(p12,p22,p2p3,x,y,z))
    collerr(p12+p22+p2p3)
    return C2_p0(x,y,z)

@loopfunction()
def D0(p1sq,p2sq,p3sq,p4sq,p12,p23,m1,m2,m3,m4):
    """ $D_{0}(p_1^2, p_2^2, p_3^2, p_4^2, (p_1 + p_2)^2, (p_2 + p_3)^2, m_1^2, m_2^2, m_3^2, m_4^2)$ wrapper for pycollier """
    if collier:
        return (pyCollier.d0(p1sq,p2sq,p3sq,p4sq,p12,p23,m1,m2,m3,m4))
    if not collier:
        raise ModuleNotFoundError('Evaluation of D0 integrals requires pyCollier!')



__doc__ = r"""# $\LaTeX$ modules
This module consists of two sub-modules:
  * anyBSM.latex.printer  
    Utilities to convert UFO parameters into valid $\LaTeX$ (in case they
    have not a meaningful `texname` attribute). Also provides a
    sympy.printing.latex.LatexPrinter-class which is used to output
    symbols in their $\LaTeX$ representation when working in interactive
    Jupyter notebooks.
  * anyBSM.latex.drawer  
    A module which draws Feynman diagrams for a given set of
    internal/external fields as well as their results.

## How do I generate the Feynman diagrams?
### With the command line
```bash
anyBSM SM -e numerical -cc -R OS --tex 0.1
```
draws all Feynman diagrams involved in the calculation of the
trilinear Higgs self coupling - including those that enter the
counterterms. On the example above, only diagrams which contribute
with more than $0.1\text{GeV}^{n}$ are drawn. The option `--tex`
without any cut-off draws all diagrams.
For each amplitude a $\LaTeX$ file is generated at
```
<Path to UFO model>/latex/<model name>_<external states>_<evaluation
mode>_results.tex
```
I.e. the above command for the SM model generates the following files:
```console
$ ls -gG SM/latex/*pdf
-rw-r--r-- 1 121936 Jul 11 15:19 SM/latex/SM_hhh_numerical_results.pdf
-rw-r--r-- 1  75855 Jul 11 15:20 SM/latex/SM_AA_numerical_results.pdf
-rw-r--r-- 1 108158 Jul 11 15:20 SM/latex/SM_AZ_numerical_results.pdf
-rw-r--r-- 1 118399 Jul 11 15:19 SM/latex/SM_hh_numerical_results.pdf
-rw-r--r-- 1 147790 Jul 11 15:19 SM/latex/SM_WpWpc_numerical_results.pdf
-rw-r--r-- 1 138098 Jul 11 15:20 SM/latex/SM_ZZ_numerical_results.pdf
```
which are the diagrams for the triple-Higgs vertex and all
selfenergies entering the on-shell counterterms.
The `.tex` files are automatically compiled using
[lualatex](https://www.luatex.org/) which is available in the
package repositories of all common operation systems. The Diagrams are
generated with the help of $\TikzFeynman$ which is shipped with the
code (no separate installation required).

$\LaTeX$ output is also supported in the analytical/abbreviation
modes:
```bash
anyBSM SM -e analytical -cc -R OS --tex
```
which places the analytic results next to the Feynman diagrams. The
file names are as in the numerical case except that they do end with `*analytical_results.pdf`.

### With the library
The `anyBSM.anyProcess.anyProcess.process` function (as well as many
user functions such as anyBSM.anyH3.anyH3.lambdahhh) has a argument
`draw=True` which enables the $\LaTeX$ generation and compilation.
```python
from anyBSM import anyBSM
SM = anyBSM('SM', evaluation = 'numerical')
SM.Sigma('Wp',draw=True)
# returns W-selfenergy and writes diagrams to ~/SM_WpWpc_numerical_results.pdf as well as the model directory
```
The `draw` argument can again be a `bool` or a `float` but also an instance of anyBSM.latex.drawer.DrawProcess 
In the latter case, diagrams are added to the list of diagrams. anyBSM.latex.drawer.DrawProcess.write needs to 
be invoked yourself (see anyBSM.latex.drawer for more information).

## Compilation failed
Most likely there is a syntax error in one of the `texname`-fields of your
UFO model. Have a look at the `.log` files in the models `latex`
directory and try to compile the `.tex` files manually with different
compilers/flags. Also make sure that
[lualatex](https://www.luatex.org/) is installed and can be found 
on your system (e.g. issue `which lualatex`).

## How do I make use of $\LaTeX$ symbols in Jupyter?
The function `anyBSM.anyModel.anyModel.sympify_parameters` initializes
all UFO parameters as sympy parameters and also generates $\LaTeX$
representation for them as well as starts a printing session. It is
automatically called in the analytic evaluation mode:
```python
from anyBSM import anyBSM
SM = anyBSM('SM', evaluation = 'analytical')
hhhcoupling = SM.sympify(SM.getcoupling('h','u3','u3bar')['L'].value)
hhhcoupling # this should render '- \\frac{\\sqrt{2} i Yu_{33}}{2}' 

from sympy import latex
latex(hhhcoupling) # returns latex string
```

If everything fails, you can create your own sympy printer like this:
```python
from anyBSM.latex.printer import anyBSMPrinter
from anyBSM.utils import import_sympy
SM.sympify_parameters() # adds LaTeX strings and sympy symbols to the dictionaries `SM.texnames` and `SM.symbols` corresponding to the UFO parameters

# define custom printer
def latex(expr, **settings):
    sett = {'symbol_names': SM.texnames}
    sett.update(settings)                                      
    return anyBSMPrinter(sett).doprint(expr)
sp = import_sympy() # make sure to import sympy only once
sp.init_printing(latex_printer=latex) # initialize sympy printing
```
"""



from __future__ import division
import logging
from tqdm.auto import tqdm
from os import path
import json
from collections import ChainMap
from hashlib import md5
from typing import Union
import cmath # noqa: F401
from cmath import pi, sin, cos, tan, atan, sqrt, log # noqa: F401
from anyBSM.loopfunctions import lnbar, A0, B0, B00, dB0, dB00, B1, dB1, C0, C1, C2, eps, set_renscale # noqa: F401
from anyBSM.ufo.function_library import complexconjugate, conjugate, Abs, I, Re, Im # noqa: F401
from anyBSM.anyProcess import anyProcess
from anyBSM.utils import import_sympy
from anyBSM.locks import CreateLock, safe_isfile

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__doc__ = """
<div class="mermaid">
flowchart RL
    subgraph classes
    anyModel --> anyProcess --> anyH3;
    end
    subgraph methods
    anyH3 --> lambdahhhCT --> lambdahhh
    end
</div>
"""

class anyH3(anyProcess):
    r""" Calculation of the trilinear Higgs self coupling in OS/$\MS$/user-defined renormalization schemes.

    This class inherits from `anyBSM.anyProcess` which itself inherits from `anyBSM.anyModel`.
    For more information on how to initialize this class, see the documentation of the mother-class `anyBSM.anyModel.anyModel.__init__`  or consult the online documentation.
    The class `anyBSM.anyH3` implements two methods:
      * `anyBSM.anyH3.anyH3.lambdahhh()`: actual computation of the full one-loop corrected trilinear self-coupling (including the renormalization-specific contributions)
      * `anyBSM.anyH3.anyH3.lambdahhhCT()`: automatic caclulation of all renormalization-specific contributions

    It is assumed that a proper renormalization scheme has already been loaded and all SM-like particles/parameters are correctly identified.
    To ensure this `anyBSM.anyModel.anyModel.load_renormalization_scheme()` should be issued beforehand (see example below).

    Example
    ```python
    from anyBSM import anyH3
    SM = anyH3('SM')
    SM.load_renormalization_scheme('OS')
    lam = SM.lambdahhh()
    lamLO = lam[0]
    lamNLO = sum(lam).real
    lamNLO/lamLO
    ```
    """
    def lambdahhh(
        self,
        fields : list = ['h_SM','h_SM','h_SM'],
        momenta: list = [0,0,0],
        only_topologies: list = [],
        simplify: bool = False,
        draw: Union[bool, int] = False,
        tadpoles: bool = None,
        wfrs: bool = None,
        exclude_particles: list = [],
        exclude_particles_pdg: list = [],
        exclude_CTs: list = [],
        parameters = {}
    ) -> dict:
        """ Wrapper around anyBSM.process and anyH3.lambdahhhCT:
            1. Ensures that the SM-like Higgs boson $h$ is correctly
            set/found in the model for the specified parameter set.
            2. Calculates all genuine tree-level and one-loop diagrams
            to $\\lambda_{hhh}$ as well as external-leg-,
            tadpole- and counter-term-contributions.
            3. Returns tuple of individual contributions

            Args:
                fields: specify external scalars
                momenta: specify external momenta
                    only_topologies: consider only diagrams from specific
                    topologies (see `anyBSM.topologies`)
                simplify: simplify results using sympy
                draw: whether to draw Feynman diagrams using
                    [tikz-feynman](https://arxiv.org/abs/1601.05437). If set to
                    a number !=0 and in evaluation mode 'numerical',
                    this number is used as a cut-off to draw only diagram which
                    contribute with an absolute value larger than the cut-off.
                tadpoles: turn on (`True`) or off (`False`) all tadpole-insertion contributions i.e. all non-PI tadpole diagrams (warning: this is
                    not physically correct and needs to be compensated by using a customCT in the schemes.yml).
                    Can also be controlled from within the schemes.yml using the `tadpole: True/False` statement.
                    Default: `None` (use schemes.yml which defaults to `True`).
                wfrs: turn on (`True`) or off (`False`) all external diagrams with external-leg corrections.
                    Similar to the tadpoles, this can be set in the schemes.yml as well.
                    Default: `None` (uses schemes.yml which defaults to `True`)
                exclude_particles: list of fields (listed by names) to exclude from diagrams -- passed on to anyBSM.process and anyH3.lambdahhhCT.
                exclude_particles_pdg: list of fields (listed by PDG numbers) to exclude from diagrams -- passed on to anyBSM.process and anyH3.lambdahhhCT.
                parameters: optionally run `anyBSM.anyModel.setparameters(parameters)` so set input parameters
                exclude_CTs: list of parameter_counterterms (defined in schemes.yml) to ignore in the calculation. This is useful when determining those CTs via OS conditions

            Return:
                dictionary of the individual contributions: total, treelevel, genuine, wfr, tadpoles, massren, vevren, customren
        """

        if len(fields) != 3:
            logger.error('`fields` arguments of lambdahhh should be a list of length 3!')
            return
        if parameters:
            self.setparameters(params=parameters)
        if not self.SM_particles['Higgs-Boson']:
            self.find_SM_particles()
        if 'h_SM' in fields and not self.SM_particles['Higgs-Boson']:
            logger.error("Cannot calculate lambda_hhh without you specifying what h to use!")
            return
        fields = [self.SM_particles.get('Higgs-Boson', 'h_SM') if type(f) is str and f == 'h_SM' else f for f in fields]
        h1, h2, h3 = self._fieldlist(fields)
        if momenta[0] == "auto":
            if self.evaluation == 'numerical':
                momenta = [h1.nmass**2, h2.nmass**2, h3.nmass**2]
            else:
                momenta = ['p12','p22','p32']

        results, drawer = self.process(h1, h2, h3,
                               momenta = momenta,
                               only_topologies = only_topologies,
                               tadpoles = tadpoles,
                               wfrs = wfrs,
                               simplify = simplify,
                               draw = draw,
                               exclude_particles = exclude_particles,
                               exclude_particles_pdg = exclude_particles_pdg,
                               return_drawer = True)

        treelevel = self._eval(f"-({results['ThreePointTree']})")
        genuine = self._eval(f"-({results['ThreePointB']} + {results['ThreePointC']})")
        wfr = self._eval(f"-({results['ThreePointWFRA']} + {results['ThreePointWFRB']})")
        tads = self._eval(f"-({results['ThreePointWFRT']} + {results['ThreePointTA']})")

        CTs = self.lambdahhhCT(
                h1, h2, h3,
                momenta = momenta,
                simplify = simplify,
                draw = False,
                only_topologies = only_topologies,
                tadpoles = tadpoles,
                exclude_particles = exclude_particles,
                exclude_particles_pdg = exclude_particles_pdg,
                exclude_CTs = exclude_CTs)

        vevren = self._eval(CTs['VEV_CT'])
        massren = self._eval(CTs['mass_CTs'])
        customren = self._eval(CTs['customCT'])

        results['ThreePointCT'] = self._eval(f'- {vevren} - {massren} - {customren}')
        if draw and drawer:
            drawer.add_diagram(
                    'ThreePointCT',
                    [],
                    results['ThreePointCT'])
            drawer.write(results)

        total = treelevel + genuine + wfr + tads + vevren + massren + customren
        return {
            'total': total,
            'treelevel': treelevel,
            'genuine': genuine,
            'wfr': wfr,
            'tads': tads,
            'massren': massren,
            'vevren': vevren,
            'customren': customren
            }

    def lambdahhhCT(
            self,
            h1, h2, h3,
            simplify: bool = False,
            momenta : list = [0,0,0],
            only_topologies: list = [],
            draw: Union[bool, int] = False,
            tadpoles: bool = None,
            exclude_particles: list = [],
            exclude_particles_pdg: list = [],
            exclude_CTs: list = []
    ) -> dict:
        """ Calculates the counter-term contribution to the trilinear
        self-coupling at the one-loop order according to the
        specification of the loaded renormalization scheme (see
        anyBSM.load_renormalization_scheme()).\n
        The calculation involves several steps:
            1. Calculate the tree-level value for $\\lambda_{h_1h_2h_3}$ and
            express it in terms of all external parameters (masses, VEVs):
            $\\lambda_{hhh}(m_{h_1}, m_{h_2}, \\dots, v_{SM},
            v_{i},\\dots$).
            2. If `VEV_counterterm: OS` was specified in the
            renormalization scheme, it calculates
            $$\\delta^{(1)}_{v_{SM}}\\lambda_{h_1h_2h_3} = \\frac{\\partial
            \\lambda_{hhh}}{\\partial v_{SM}}\\delta^{(1)}v_{SM}$$
            with
            $$\\delta^{(1)}v_{SM} = v_{SM}\\left(\\frac{\\delta^{(1)}M_W^2}{M_W^2} - \\frac{\\delta^{(1)}M_Z^2}{2M_Z^2} -\\frac{\\delta^{(1)}e}{e} - \\frac{\\delta^{(1)}M_W^2}{2\\sin^2\\theta_w} + \\frac{\\delta^{(1)}M_Z^2}{2\\sin^2\\theta_w}\\right)$$
            where the OS counter-terms for the W- and Z-boson as well as the
            electric charge are extracted from the transverse parts of the
            W-, Z-,  Photon- and Z-Photon-selfenergies. See [2305.03015](https://arxiv.org/abs/2305.03015) for more details.
            3. For all `'OS'`-entries `mass_counterterms: {'hi': 'OS',
            ...}` the contribution
            $$\\delta^{(1)}_{m_{h_i}}\\lambda_{hhh} = -\\frac{\\partial
            \\lambda_{hhh}}{\\partial m_{h_i}}\\frac{1}{2
            m_{h_i}}\\mathrm{Re}\\Sigma^{(1)}_{h_i}(p^2=m_{h_i}^2)$$
            is calculated.
            4. If a `custom_counterterm` with `process: [h1, h2, h3]` has been specified in the renormalization
            scheme, the supplied `condition` is being evaluated. For instance
            a valid value could look like:

                 <!-- language: yml -->

                    [...]
                    wfrs: false
                    parameter_counterterms:
                        - parameter: betaH
                          counterterm: dbetaH
                          condition: (Re(Sigma('Hm1','Hm2',momentum='0')) + Re(Sigma('Hm1','Hm2',momentum='MHm2**2')))/(2*MHm2**2)
                    custom_counterterms:
                        - process: [h1, h1, h1]
                          condition: |
                            dZhh = Sigmaprime('h', momentum='Mh**2')
                            customCT = 3*Mh**2/vvSM*3/2*dZhh"

             the value of `customCT` **must** be set in order to
            add the contribution to the final result:
            $$\\delta^{(1)}_{\\text{custom}}\\lambda_{h_1h_2h_3} = {\\tt \\text{customCT}} + \\sum_{x}^{\\tt \\text{parameter_counterterms}} \\frac{\\partial \\lambda_{h_1 h_2 h_3}}{\\partial x} \\delta x$$
            The `customCT` provides access to the full `self` object of
            the model with all its methods such as `anyBSM.Derivative`,
            `anyBSM.Sigma`, `anyBSM.Sigmaprime` or `anyBSM.process. This is
            particularly useful for more complex renormalization schemes. Note that specific choices of momenta, or restrictions on topologies, particles, must be included in the customCT definition. Contributions from defined `parameter_counterters` via the vertex counterterms of the considered amplitude are taken into account automatically and must not be specified in the `custom_counterterms` (their definitions can, however, be reused therein).

        Args:
            simplify:  simplify results using sympy
            only_topologies: consider only diagrams from specific
                topologies (see `anyBSM.topologies`)
            draw: whether to draw Feynman diagrams using
                [tikz-feynman](https://arxiv.org/abs/1601.05437). If set to
                a number !=0 and in evaluation mode 'numerical',
                this number is used as a cut-off to draw only diagram which
                contribute with an absolute value larger than the cut-off.
            tadpoles: turn off all tadpole-insertion contributions (warning: this is
                not physically correct unless a customCT with the correct tadpole treatment is added!)
            exclude_particles: list of fields (listed by names) to exclude from diagrams -- passed on to anyBSM.Sigma and anyBSM.Sigmaprime via sigmaargs.
            exclude_particles_pdg: list of fields (listed by PDG numbers) to exclude from diagrams -- passed on to anyBSM.Sigma and anyBSM.Sigmaprime via sigmaargs.
            exclude_CTs: list of parameter_counterterms (defined in schemes.yml) to ignore in the calculation. This is useful when determining those CTs via OS conditions

        Returns:
            A dictionary `{`$\\text{'VEV_CT'}:
            \\delta^{(1)}_{v_{SM}}\\lambda_{hhh},\\, \\text{'mass_CTs'}:
            \\sum_{i} \\delta^{(1)}_{m_{h_i}}\\lambda_{hhh},\\,
            \\text{'customCT'}: \\delta^{(1)}_{\\text{custom}}\\lambda_{hhh}$ `}`
        """

        if not self.scheme:
            logger.warning("no renormalization scheme defined. Assuming all input parameters are MSbar values.")
            return {'VEV_CT': '', 'mass_CTs': '', 'custom_CT': ''}
        argstring = f'Topologies: {only_topologies}; Particle exclusions: {exclude_particles}, {exclude_particles_pdg}; Tadpoles: {tadpoles}; Excluded CTs: {exclude_CTs}'
        argshash = md5(argstring.encode('utf-8')).hexdigest()
        CTfile = path.join(self.cachedir, f'result_{h1}{h2}{h3}CT_{self.scheme_name}_scheme_{argshash}.json')
        momenta_d = {f'PEXT{i+1}2' : m for i,m in enumerate(momenta)}
        if self.caching > 1 and safe_isfile(CTfile):
            with open(CTfile, 'r') as f:
                logger.info(f'reading counterterm contributions for "{self.scheme_name}"-scheme from cache ({CTfile})')
                result = json.load(f)
            if self.evaluation == 'numerical':
                momenta_d = {k: self._eval(v) for k,v, in momenta_d.items()}
                for k,v in result.items():
                    try:
                        result[k] = self._eval(v, **momenta_d)
                    except Exception as e:
                        raise Exception(f"There was an error when evaluating the '{k}'-counterterm contribution") from e

            elif self.evaluation == 'analytical':
                for r in result:
                    for m,md in momenta_d.items():
                        result[r] = result[r].replace(m, str(md))
                    for name,c in reversed(self.couplings.items()):
                        result[r] = result[r].replace(name, f'({c.value})')
            return result
        firstrun = self.caching > 1 and self.evaluation != 'abbreviation' and not only_topologies and not draw
        if firstrun:
            evalSAVE = self.evaluation
            logger.info('No CT results found in cache. Calculating in abbreviation-mode to fill cache.')
            self.set_evaluation_mode('abbreviations')
            lock = CreateLock(CTfile)

        # options used for calculation of all two-point funtions in the following
        sigmaargs = {
                'momentum': 'auto' if self.OSmomenta else 0,
                'only_topologies': only_topologies,
                'tadpoles': tadpoles,
                'simplify': simplify,
                'draw': draw,
                'exclude_particles': exclude_particles,
                'exclude_particles_pdg': exclude_particles_pdg}

        OSlist = [self.all_particles[k] for k,v in self.scheme['mass_counterterms'].items() if v == 'OS']
        treelevel_needed = OSlist or self.scheme['VEV_counterterm'] == 'OS' or self.scheme['custom_counterterms'] or self.scheme['parameter_counterterms']
        # calculate tree-level analytical (if not already)
        if treelevel_needed and ((h1,h2,h3) not in self.treelevel or self.evaluation != 'analytical'):
            logger.info('calculate counterterm contributions to lambda_hhh')
            evalSAVE2 = str(self.evaluation)
            self.set_evaluation_mode('analytical') # enforce analytic evaluation to be able to perform derivatives

            logger.info('Calculating tree-level value of lambda_hhh and expressing it in terms of all input parameters.')
            self.treelevel[(h1,h2,h3)] = self.process(h1, h2, h3, momenta = [], only_topologies = ['ThreePointTree'], simplify = simplify)['ThreePointTree']
            self.treelevel[(h1,h2,h3)] = self.treelevel[(h1,h2,h3)].replace('cmath.','')

            treemasses = [p.mass.name for p in OSlist]

            self.treelevel[(h1,h2,h3)] = -1*self.SolveDependencies(self.treelevel[(h1,h2,h3)], simplify = simplify, exclude = treemasses+[self.SM_parameters['VEV'].name])
            logger.info(f'lamda_hhh^tree = {self.treelevel[(h1,h2,h3)]}')

            treemasses = [m for m in treemasses if m in str(self.treelevel[(h1,h2,h3)])]
            logger.info(f"found the following scalar masses appearing at the tree-level: {treemasses}")

            self.set_evaluation_mode(evalSAVE2)

        # contributions from mass renormalization
        OSliststr = [f"{p.name} ({p.mass})" for p in OSlist]
        logger.info(f"the masses of the following particles are renormalized OS: {' '.join(OSliststr) if OSlist else 'none'}")
        massren = self._eval('')
        for p in tqdm(OSlist, leave = False, disable = not self.progress, desc = 'calculate mass CT contributions'):
            logger.info(f'taking derivative of tree-level result w.r.t. {p.mass.name}')
            massderivative = self.Derivative(self.treelevel[(h1,h2,h3)], p.mass.name, simplify = simplify, dependencies = False)
            if massderivative == '0' or not massderivative:
                logger.warning(f'derivative of lambda_hhh^tree w.r.t {p.mass.name} vanishes!')
                continue
            selfenergy = self.Sigma(p.name, **sigmaargs)
            massren += self._eval(f"-({massderivative})*Re({selfenergy})/(2*{p.mass})")

        # contributions from vev renormalization
        vevren = self._eval('')
        vevderivative = '0'
        if self.scheme['VEV_counterterm'] == 'OS':
            logger.info('calculate CT contribution due to electroweak VEV')
            logger.info(f'taking derivative of tree-level result w.r.t. {self.SM_parameters["VEV"].name}')
            vevderivative = self.Derivative(self.treelevel[(h1,h2,h3)], self.SM_parameters['VEV'].name, simplify = simplify, dependencies = False)
            if str(vevderivative) == '0':
                logger.warning('derivative of lambda_hhh^tree w.r.t the SM VEV vanishes!')

        if str(vevderivative) != '0':
            # parameters
            MW2 = self.getmass('W-Boson', 2)
            MZ2 = self.getmass('Z-Boson', 2)
            CW = f'sqrt(({MW2})/({MZ2}))'
            SW = f'(sqrt(1-({MW2})/({MZ2})))'
            alphaQEDinverse = f"1/({self.SM_parameters['alphaQEDinverse']})"
            EL = f"(2*sqrt({alphaQEDinverse})*sqrt(pi))"
            VEV = f'2*sqrt(({MW2}*{MZ2}-{MW2}**2)/{MZ2})/{EL}'
            dalpha = self.SM_parameters['Dalpha'].name

            # W/Z selfenergies
            SigmaW = self.Sigma(self.all_particles['W-Boson'], **sigmaargs)
            SigmaZ = self.Sigma(self.all_particles['Z-Boson'], **sigmaargs)

            # Photon(-Z-mixing) selefenergies
            SigmapAnolf = self.Sigmaprime(
                    self.all_particles['Photon'],
                    **ChainMap({'momentum': 0, 'exclude_particles_pdg': list(set(exclude_particles_pdg) | set(self.light_SM_fields))}, sigmaargs))
            SigmaAlf = self.Sigma(
                    self.all_particles['Photon'],
                    **ChainMap({'momentum': MZ2, 'exclude_particles_pdg': list(set(exclude_particles_pdg) | set(self.heavy_fields))}, sigmaargs))
            SigmaAZ = self.Sigma(
                    self.all_particles['Photon'], self.all_particles['Z-Boson'],
                    **ChainMap({'momentum': 0}, sigmaargs))

            # counterterms
            deltaMW = f'Re({SigmaW})/({MW2})'
            deltaMZ = f'Re({SigmaZ})/({MZ2})'
            deltaEL = f'(1/2*(({dalpha}) + Re({SigmapAnolf}) + Re({SigmaAlf})/({MZ2}))\
                    +  SignSinThetaW*({SW}/({CW}*{MZ2}))*Re({SigmaAZ}))'
            vevCT = f'{VEV}*(({deltaMW})/2 - {deltaEL} +\
                    ({CW}**2)/(2*{SW}**2)*(({deltaMZ}) - ({deltaMW})))'

            vevren += self._eval(f'({vevCT})*({vevderivative})')

        if self.scheme['custom_counterterms'] or self.scheme['parameter_counterterms'] or self.scheme.get('custom_CT_hhh', False):
            # define a few convenient functions which can directly be used in the custom_CT_hhh string
            Tadpole = self.Tadpole           # noqa: F841
            Sigmaprime = self.Sigmaprime     # noqa: F841
            Sigma = self.Sigma               # noqa: F841
            Derivative = self.Derivative     # noqa: F841
            process = self.process           # noqa: F841
            lambdahhh = self.lambdahhh       # noqa: F841
            sympify = self.sympify       # noqa: F841
            getcoupling = self.getcoupling   # noqa: F841
            SM_particles = self.SM_particles # noqa: F841
            treelevel = self.treelevel[(h1,h2,h3)]  # noqa: F841
            lambdahhh_tree = treelevel  # noqa: F841 # v1.0 legacy
            self.sympify_parameters()
            self.symbols['Derivative'] = Derivative
            self.symbols['getcoupling'] = getcoupling
            self.symbols['Tadpole'] = lambda *x,**y: sympify(Tadpole(*x,**y))
            self.symbols['Sigma'] = lambda *x,**y: sympify(Sigma(*x,**y))
            self.symbols['Sigmaprime'] = lambda *x,**y: sympify(Sigmaprime(*x,**y))
            self.symbols['process'] = process
            self.symbols['lambdahhh'] = lambdahhh
            sp = import_sympy()

        customren = self._eval('')

        # sympify user defined parameter counterterms
        self.parameter_counterterms = {}
        all_cts = {}
        for pc in self.scheme['parameter_counterterms']:
            if 'parameter' not in pc:
                logger.error(f'No UFO "parameter" defined for counterterm: {pc}')
                continue
            if pc['counterterm'] in exclude_CTs or pc['parameter'] in exclude_CTs:
                logger.debug(f'skipping CT for {pc}')
                continue
            if pc.get('warn', True):
                param = self.parameters.get(pc['parameter'], None)
                if not param:
                    logger.error(f"Parameter {pc['parameter']} not found in the UFO model!")
                if param and param.nature != 'external':
                    logger.warning((f"Can't automatically take into account the counterterm of a parameter ({param.name}) "
                                    "which is not defined as UFO input parameter."))
                    logger.warning(("It's contribution needs to be added manually using a custom counterterm."
                                    "It is possible to reuse it's definition therein."
                                    "If you already do so, you can disable this warning by setting `warn: False` for this counterterm."))
            ctname = pc.get('counterterm', None)
            if not ctname: # automatically assing CT name if no one was defined
                ctname = 'd' + pc['parameter']
            if ctname in self.parameters: # avoid conflicts with other UFO parameters
                logger.warning(f'counterterm name {ctname} conflicts with already defined UFO parameter!')
                ctname = 'deltaCT' + pc['parameter']
            ct = pc.get('condition', None)
            if not ct:
                logger.error(f"No counterterm condition defined for {pc['parameter']} ({ctname})")
                continue
            self.symbols[ctname] = sp.Symbol(ctname)
            self.symbols_values[ctname] = self.SolveDependencies(ct, simplify = simplify)
            all_cts[ctname] = self._eval(str(self.symbols_values[ctname]), **all_cts)
            logger.info(f'caclulating contrubution from custom counterterm {ctname} onto {(h1,h2,h3)}-amplitude')
            ct_contrib = self.Derivative(self.treelevel[(h1,h2,h3)], pc['parameter'], simplify = False)*self.symbols_values[ctname]
            if simplify:
                ct_contrib = ct_contrib.simplify()
            if ct_contrib == 0 and pc.get('warn', True):
                logger.warning(f"Counterterm contribution of {ctname} to {(h1,h2,h3)} vanishes (set `warn=False` for this CT to disable this message)")
            logging.debug(f'result: {ct_contrib}')
            customren += self._eval(str(ct_contrib), **all_cts)

        proc = sorted([h1.name,h2.name,h3.name])

        # legacy of v1.0 notation for specifying CTs
        old_ct = self.scheme.get('custom_CT_hhh', None)
        if old_ct and proc not in [c['amplitude'] for c in self.scheme['custom_counterterms']]:
            logger.warning('The "cumston_CT_hhh" notation is deprecated! converting to the new convention...')
            self.scheme['custom_counterterms'].append({'amplitude': proc, 'condition': old_ct})

        # contrubtions from user defined vertex counterterms
        for ct in self.scheme['custom_counterterms']:
            amplitude = sorted(ct.get('amplitude',[]))
            if not amplitude:
                logger.error(f'Need to provide "amplitude" for custom CT. E.g. amplitude: [h1,h1,h1] (CT: {ct})"')
                continue
            if amplitude != proc:
                continue
            custom_ct_str = ct.get('condition', None)
            if not custom_ct_str:
                logger.error(f'No "condition" provided for {amplitude} counterterm')
                continue

            logger.info('calculate custom CT contribution to lambda_hhh')
            CT = ''
            # prepare a local environment with sympy expressions for the evaluation of custom counterterms
            locs = {}
            locs.update(locals())
            locs.update(self.symbols_values)
            locs.update(self.symbols)

            try:
                exec(custom_ct_str, globals(), locs)
            except Exception as e:
                raise Exception("CounterTermError", "could not execute your custom counter term due to following error: \n" + str(e))
            CT = locs.get('customCT', '')
            # return value of v1.0 legacy
            if hasattr(self,'custom_CT_hhh') and CT == '':
                CT = self.custom_CT_hhh

            if CT == '':
                logger.error('your custom counterterm did not store a condition in the variable `CT`')
            try:
                CT = self._eval(str(CT))
            except Exception as e:
                logger.error("could not calculate custom counter term due to following error: \n" + str(e))
                CT = ''
            if CT == 0:
                logger.warning('your custom counter term evaluated to zero (0)!')
            customren += CT

        result = {'VEV_CT': vevren, 'mass_CTs': massren, 'customCT': customren}

        if not only_topologies and self.evaluation != 'numerical':
            logger.debug(f'Writing results to cache ({CTfile}).')
            with open(CTfile, 'w') as f:
                json.dump(result, f)
            lock.release()
        if firstrun:
            self.set_evaluation_mode(evalSAVE)
            if self.evaluation == 'numerical':
                momenta_d = {k: self._eval(v) for k,v, in momenta_d.items()}
                for k,v in result.items():
                    try:
                        result[k] = self._eval(v, **momenta_d)
                    except Exception as e:
                        raise Exception(f"There was an error when evaluating the '{k}'-counterterm contribution") from e

            elif self.evaluation == 'analytical':
                for r in result:
                    for m,md in momenta_d.items():
                        result[r] = result[r].replace(m, str(md))
                    for name,c in reversed(self.couplings.items()):
                        result[r] = result[r].replace(name, f'({c.value})')

        return result



from .anyBSM import anyBSM
from .anyModel import anyModel
from .anyProcess import anyProcess
from .anyEWPO import anyEWPO
from .anyH3 import anyH3
from .anyPerturbativeUnitarity import anyPerturbativeUnitarity
from .anyEffectiveCoupling import anyEffectiveCoupling

__doc__ = r""" The anyBSM library

This is the anyBSM python package.  
The library is organized in various (sub)modules which define physics obserables (and helper-classes), such as e.g. `anyBSM.anyH3`.
The module `anyBSM.anyBSM` inherits from those modules and makes all observables accessible within one class:
```python
from anyBSM import anyBSM
SM = anyBSM('SM')
# this object can call functions from all sub-modules
SM.lambdahhh() # defined in the `anyBSM.anyH3`-module
SM.eigSSSS() # defined in the `anyBSM.anyPerturbativeUnitarity`-module
SM.getcoupling('h','Z','Z') # defined in the root module `anyBSM.anyModel`
```
<div class="mermaid">
flowchart RL

  subgraph user[main classes]
    direction TB
    anyModel --> anyProcess
    anyProcess --> anyH3
    anyProcess --> anyEWPO
    anyProcess --> ...
    anyH3 --> anyEffectiveCoupling
    anyH3 --> anyHH
    anyH3 --> anyBSM
    ... --> anyBSM
  end
  subgraph internal[internal modules]
    direction TB
    A[topologies<br>diagrams<br>loopfunctions]
  end
  subgraph external[external optional modules]
    direction TB
    pyCollier
  end
  internal --- user
  external --- user
</div>
"""



from __future__ import division
import logging
from tqdm.auto import tqdm
from math import factorial
from os import path
import json
from collections import defaultdict, OrderedDict
from itertools import permutations, combinations_with_replacement
from hashlib import md5
from typing import Union
import cmath # noqa: F401
from cmath import pi, sin, cos, tan, atan, sqrt, log # noqa: F401
from anyBSM.loopfunctions import lnbar, A0, B0, B00, dB0, dB00, B1, dB1, C0, C1, C2, eps, set_renscale # noqa: F401
from anyBSM.ufo.function_library import complexconjugate, conjugate, Abs, I, Re, Im # noqa: F401
from anyBSM.topologies import TOPOLOGIES
from anyBSM.diagrams import GenericDiagrams
from anyBSM.utils import fieldtypes
from anyBSM.locks import safe_isfile, CreateLock
from anyBSM.anyModel import anyModel

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
__doc__ = """
<div class="mermaid">
flowchart RL
    subgraph classes
    anyModel --> anyProcess;
    end
    subgraph methods
    anyProcess --> insert_fields --> process
    end
    subgraph wrappers/helpers
    process --> Sigma
    process --> Sigmaprime
    process --> Tadpole
    end
</div>
"""

class anyProcess(anyModel):
    """ Class for the generation and caclulation of all Feynman diagrams for a given set of external fields.
    This class inherits from `anyBSM.anyModel`.
    For more information on how to initialize this class, see the documentation of the mother-class `anyBSM.anyModel.anyModel.__init__`  or consult the online documentation.
    """

    def process(
            self,
            *externalfields,
            momenta: list = [],
            derivative: bool = False,
            only_topologies: list = [],
            tadpoles: bool = None,
            wfrs: bool = None,
            simplify: bool = False,
            draw: Union[bool, int] = False,
            export: str = '',
            exclude_particles: list = [],
            exclude_particles_pdg: list = [],
            compute_permutations : bool = True,
            return_drawer: bool = False
    ) -> defaultdict:
        """ For given external fields `externalfields` calculate all possible one-loop diagrams.

        The procedure is roughly as follows:
          * find all possible insertions for all possible topologies using `anyProcess.insert_fields`
          * iterate over the resulting diagram insertions
          * extract the appropriate couplings and masses specified in the UFO for the calculation of the diagrams
          * use generic results for those diagrams; apply color factors if necessary

        Args:
            externalfields: list of field names or objects
            momenta: **squared** momenta
            derivative: whether to calculate the derivate w.r.t
                squared external momentum (only for two-point functions)
            only_topologies: list of specific topologies of
                diagrams to restrict to
            tadpoles: turn on (`True`) or off (`False`) all tadpole-insertion contributions i.e. all non-PI tadpole diagrams.
                Can also be controlled from within the schemes.yml using the `tadpole: True/False` statement.
                Default: `None` (use schemes.yml which defaults to `True`).
                Note that `only_topologies` is superior to `tadpoles` and `tadpoles` is superior to the value in the schemes.yml.
            wfrs: turn on (`True`) or off (`False`) all external diagrams with external-leg corrections.
                Similar to the tadpoles, this can be set in the schemes.yml as well.
                Default: `None` (uses schemes.yml which defaults to `True`)
                Note that `only_topologies` is superior to `wfrs` and `wfrs` is superior to the value in the schemes.yml.
            simplify: whether to simplify results using sympy
            export: save results to a file (absolute path). If the
                file exists, the result is appended to that file.
            draw: whether to draw Feynman diagrams using
                [tikz-feynman](https://arxiv.org/abs/1601.05437). If set to
                a number !=0 and in evaluation mode 'numerical',
                this number is used as a cut-off to draw only diagram which
                contribute with an absolute value larger than the cut-off.
            exclude_particles: list of fields to exclude from diagrams (listed by names).
            exclude_particles_pdg: list of fields to exclude from diagrams (listed by PDG numbers).

        Returns:
            analytical/numerical/abbreviated result depending on the evaluation mode
            `anyModel.evaluation`.
        """
        externalfields = self._fieldlist(*externalfields)
        extfields = ''.join([f.name for f in externalfields])
        nextfields = len(externalfields)
        momenta = [0 for i in range(nextfields)] if not momenta else momenta
        momenta_placeholder = [f'PEXT{i+1}2' for i in range(nextfields)]
        derivative_str = '' if not derivative else '_derivative'
        wfrs_str = self.scheme.get('wfrs', True) if wfrs is None else wfrs
        tadpoles_str = self.scheme.get('tadpoles', True) if tadpoles is None else tadpoles
        argstring = f'Topologies: {only_topologies or "all"}; Particle exclusions: {exclude_particles + exclude_particles_pdg or "none"}; Simplify: {simplify}; WFRs: {wfrs_str}; Tadpoles: {tadpoles_str}; Permutations: {compute_permutations};'
        # TODO: write argstring as comment to the json file
        argshash = md5(argstring.encode('utf-8')).hexdigest()
        cachefile = path.join(self.cachedir, f'result_{extfields}{derivative_str}_{argshash}.json')
        loop_particles = [p for p in self.loop_particles if (p.name not in exclude_particles and abs(p.pdg_code) not in exclude_particles_pdg)]
        loop_particles = loop_particles + [p.anti() for p in loop_particles if p != p.anti()]
        zero = self._eval('')
        drawer = False
        if self.caching > 1 and not draw and safe_isfile(cachefile):
            logger.info(f'reading results for {extfields}-amplitude from cache ({cachefile}).')
            with open(cachefile, 'r') as f:
                result = defaultdict(lambda *x: zero)
                result.update(json.load(f))
                result.pop('setup', None)
        else:
            firstrun = self.caching > 1 and self.evaluation != 'abbreviation' and not draw
            if firstrun:
                lock = CreateLock(cachefile)
                evalSAVE = self.evaluation
                logger.info('No results found in cache. Calculating in abbreviation-mode to fill cache.')
                self.set_evaluation_mode('abbreviations')
                zero = self._eval('')
            if self.evaluation == 'numerical':
                momenta_in = [self._eval(m) for m in momenta]
            elif compute_permutations:
                momenta_in = momenta_placeholder
            else:
                momenta_in = momenta
            momenta_in_save = momenta_in
            if draw.__class__.__name__ == 'DrawProcess':
                drawer = draw
                draw = True
            elif draw:
                from anyBSM.latex.drawer import DrawProcess # import here to avoid sympy import in numerical runs
                drawer = DrawProcess(self, externalfields, momenta_in, simplify, draw, process_options = argstring)
            else:
                drawer = False
            all_insertions = OrderedDict()
            if only_topologies:
                for topo in only_topologies:
                    all_insertions.update(self.insert_fields(
                        *externalfields,
                        only_topology=topo,
                        wfrs = wfrs_str,
                        tadpoles = tadpoles_str,
                        exclude_particles = exclude_particles,
                        exclude_particles_pdg = exclude_particles_pdg))
            else:
                all_insertions. update(self.insert_fields(
                    *externalfields,
                    wfrs = wfrs_str,
                    tadpoles = tadpoles_str,
                    exclude_particles = exclude_particles,
                    exclude_particles_pdg = exclude_particles_pdg))
            diagramtype = fieldtypes(list(externalfields))
            result = defaultdict(lambda *x: zero)
            genericdiags  = GenericDiagrams(evaluation = self.evaluation, MSDR = int(not self.dimensional_reduction))
            desc = f'Calculate all diagrams for {extfields}-amplitude'
            logger.info(desc)
            bar = tqdm(all_insertions.items(), disable = not self.progress, desc = desc, leave = False, total = len(all_insertions))
            for topo, insertions in bar:
                bar.set_description(f'{desc} ({topo})')
                if topo not in TOPOLOGIES[nextfields]:
                    continue
                if 'WFR' in topo and wfrs_str == 'OS':
                    momenta_in = [self._eval(f'{p.mass.name}**2') for p in externalfields]
                else:
                    momenta_in = momenta_in_save
                topology = TOPOLOGIES[nextfields][topo]
                field_mom = [(f,m) for f,m in zip(externalfields,momenta_in)]
                field_perms = set(tuple(field_mom[i-j] for i in range(nextfields)) for j in range(nextfields)) if nextfields > 2 else []
                if topology.symmetryfactor != 1 and compute_permutations and len(field_perms) > 1:
                    logger.debug(f'Calculating field permutations: {field_perms}')
                    for perm in field_perms:
                        field_perm = [f[0] for f in perm]
                        mom_perm = [f[1] for f in perm]
                        logger.debug(f'permutation: {field_perm}, {mom_perm}')
                        tmp = self.process(
                                *field_perm,
                                momenta = mom_perm,
                                derivative = derivative,
                                tadpoles = tadpoles_str,
                                wfrs = wfrs_str,
                                simplify = simplify,
                                draw = drawer,
                                export = False,
                                exclude_particles = exclude_particles,
                                exclude_particles_pdg = exclude_particles_pdg,
                                only_topologies=[topo],
                                compute_permutations=False,
                                return_drawer=True)[0]
                        if topo in tmp:
                            res = self._eval(f'+({tmp[topo]})')
                            logger.debug(f'result for this permutation: {res}')
                            result[topo] += res
                    continue

                if not hasattr(genericdiags, topology.Amplitude):
                    logger.error(f'Topology {topology.Amplitude} not found')
                    continue
                genamp = genericdiags.__getattribute__(topology.Amplitude)
                if genamp is None:
                    logger.error(f'Generic amplitude not implemented for topology {topo} for external fields = {diagramtype}')
                    continue
                i = 0
                bar2 = tqdm(insertions, leave=False, disable = not self.progress, total = len(insertions), desc = 'diagrams: ')
                for particles_s, couplings_s in bar2:
                    i += 1
                    particles = [self.all_particles[p] for p in particles_s]
                    couplings = [{k : self.couplings.get(v,v) for k,v in c.items()} for c in couplings_s[0]]
                    conj_couplings = [{k : self.couplings.get(v,v) for k,v in c.items()} for c in couplings_s[1]]
                    if any([p not in loop_particles and p.antiname not in loop_particles for p in particles]):
                        continue

                    chargefac = 1 # instead, we sum over all (anti)-particles
                    colorfac = self.ColorFactor(externalfields, particles, topo)
                    fac = self._eval(f'{colorfac}*{chargefac}/(16*pi**2)')
                    if topology.symmetryfactor != 1 and compute_permutations and len(field_perms) == 1:
                        fac = self._eval(f'{fac}*{topology.symmetryfactor}')
                    ampargs = {
                            'external': externalfields,
                            'internal': particles,
                            'coupling_objects': couplings,
                            'conj_coupling_objects': conj_couplings,
                            'derivative': derivative,
                            'factor': fac
                            }
                    tmp = genamp(momenta = momenta_in, **ampargs)

                    if tmp and simplify and self.evaluation != 'numerical':
                        try:
                            bar2.set_postfix_str('(simplifying...)')
                            tmp = self.sympify(str(tmp)+'+0', simplify = simplify)
                            bar2.set_postfix_str('')
                        except Exception as e:
                            bar2.set_postfix_str('')
                            logger.exception('Not able to simplify analytical results: ' + str(e))
                            logger.error('Continuing with un-simplified result.')
                    result[topo] += self._eval(str(tmp) + '+0')
                    if draw:
                        drawer.add_diagram(
                                topo,
                                particles,
                                tmp,
                                externalfields = externalfields,
                                momenta = momenta_in if not compute_permutations else [])

            try:
                bar.close()
                bar2.close()
            except UnboundLocalError:
                pass
            if draw and draw.__class__.__name__ != 'DrawProcess' and not return_drawer and drawer:
                drawer.write(result)

            if firstrun and self.evaluation == 'abbreviations':
                logger.debug(f'Writing results to cache ({cachefile}).')
                result['setup'] = argstring
                with open(cachefile, 'w') as f:
                    json.dump(result,f)
                result.pop('setup')
                lock.release()
                self.set_evaluation_mode(evalSAVE)

        momenta_d = {f'PEXT{i+1}2' : m for i,m in enumerate(momenta)}
        if self.evaluation == 'numerical':
            momenta_d = {k: self._eval(v) for k,v, in momenta_d.items()}
            result.update({k: self._eval(v, **momenta_d) for k,v in result.items()})
        elif compute_permutations:
            for r in result:
                for m,md in momenta_d.items():
                    result[r] = result[r].replace(m, str(md))
                if self.evaluation == 'analytical':
                    for name,c in reversed(self.couplings.items()):
                        result[r] = result[r].replace(name, f'({c.value})')

        if export:
            export = path.abspath(export)
            if not path.isdir(path.dirname(export)):
                logger.error(f'{export}: not a directory')
            else:
                with open(export, 'w') as e:
                    self._print(f'Saving results to {export}.')
                    json.dump(result, e)
        if return_drawer:
            return result, drawer
        return result

    def Sigma(self, f1, f2="", momentum: Union[str, float] = 'auto', signconvention = 1, **kwargs):
        """ Calculates the self-energy for field with name `f1`.
        This is a wrapper around `anyProcess.process()`.

        Args:
            f1: field name or object
            f2: field name or object (default: `f2=f1`)
            momentum: external momentum **squared**. 'auto' uses
                geometric mean of the squared masses of the particles.
            signconvention: flag allowing to switch between overall sign convention for self-energies. The default value, 1, corresponds to the convention employed in SARAH and many works on generic calculations, where mass counterterms are equal to *minus* the self-energy. Setting this flag to 0 changes to the convention used e.g. in FeynHiggs, with mass counterterms equal to the self-energy with a *plus*.
            **kwargs: additional keyword arguments passed to `self.process`

        Returns:
            analytical/numerical/abbreviated result depending on the evaluation mode
            `anyModel.evaluation`.
        """
        if not f2:
            f2 = f1
        p1 = self.all_particles.get(f1, f1)
        p2 = self.all_particles.get(f2, f2)
        if p1.charge == p2.charge and p2.anti() != p2: # make sure we call e.g. Hp-Hpc
            p2 = p2.anti()
        if not (p1 and p2):
            logger.error(f"invalid input for Sigma({f1},{f2}): field name(s) not found.")
            return 0
        if momentum == 'auto':
            if self.evaluation == 'numerical':
                s = (p1.nmass**2 + p2.nmass**2)/2
            elif p1.mass == p2.mass:
                s = f'{p1.mass.name}**2'
            else:
                s = f'({p1.mass.name}**2 + {p2.mass.name}**2)/2'
        else:
            s = momentum
        sigma = self.process(p1, p2, momenta = [s], **kwargs)
        sigma = self._eval('+'.join([str(r) for r in sigma.values()]))
        if not sigma:
            return 0
        else:
            if signconvention:
                return sigma
            else:
                return self._eval(f'-1*({sigma})')

    def Sigmaprime(self, *args, **kwargs):
        """ Passes all its arguments to `anyProcess.Sigma()` with
        `derivative=True` i.e. calculates the derivate of the
        two-point function w.r.t. the squared external momentum

        Returns:
            analytical/numerical/abbreviated result depending on the evaluation mode
            `anyModel.evaluation`.
        """
        return self.Sigma(*args, derivative = True, **kwargs)

    def Tadpole(self, f, **kwargs):
        """ Calculates the (scalar) tadpole diaram for field with name `f`.
        This is a wrapper around `anyProcess.process()`.

        Args:
            f1: field name (str) or UFO object
            **kwargs: additional keyword arguments passed to `self.process`

        Returns:
            analytical/numerical/abbreviated result depending on the evaluation mode
            `anyModel.evaluation`.
        """
        p = self.all_particles.get(f, f)
        if not p:
            logger.error(f"invalid input for Tadpole({f}): field not found.")
            return 0
        tad = self.process(p, **kwargs)
        return self._eval('+'.join([str(r) for r in tad.values()]))

    def insert_fields(
            self,
            *externalfields,
            only_topology: Union[None, str] = None,
            wfrs: bool = None,
            tadpoles: bool = None,
            exclude_particles: list = [],
            exclude_particles_pdg: list = []
    ) -> OrderedDict:
        """ Insert all possible field permuations for given process
        with `externalfields` into `anyBSM.topologies.TOPOLOGIES`.
        The insertions are carried out using a brute-force method which inserts
        all possible field combinations into all possible topologies and checks
        whether the resulting couplings are all existing in the UFO model.

        Args:
            externalfields: field names or objects
            only_topology: consider only this specific topology rather
                than all possible ones (see `anyBSM.topologies`).
            tadpoles: turn on (`True`) or off (`False`) all tadpole insertions i.e. all non-PI tadpole diagrams.
                Can also be controlled from within the schemes.yml using the `tadpole: True/False` statement.
                Default: `None` (use schemes.yml which defaults to `True`).
                Note that `only_topologies` is superior to `tadpoles` and `tadpoles` is superior to the value in the schemes.yml.
            wfrs: turn on (`True`) or off (`False`) all external diagrams with external-leg corrections.
                Similar to the tadpoles, this can be set in the schemes.yml as well.
                Default: `None` (uses schemes.yml which defaults to `True`)
                Note that `only_topologies` is superior to `wfrs` and `wfrs` is superior to the value in the schemes.yml.
            exclude_particles: list of fields to exclude from diagrams (listed by names).
            exclude_particles_pdg: list of fields to exclude from diagrams (listed by PDG numbers).

        Returns:
            `OrderedDict` of insertions.
        """
        topologies = TOPOLOGIES[len(externalfields)]
        tadpoles_str = self.scheme.get('tadpoles', True) if tadpoles is None else tadpoles
        if not only_topology:
            if not tadpoles_str:
                topologies = {k:v for k,v in topologies.items() if v.tadpole is False}
                logger.info('Ignoring non-1PI tadpole diagrams.')
            if wfrs is None:
                wfrs = self.scheme.get('wfrs', True)
            if not wfrs:
                logger.info('Ignoring non-1PI WFR diagrams.')
                topologies = {k:v for k,v in topologies.items() if v.wfr is False}

        insertions = OrderedDict()
        extfields = ''.join([f.name for f in externalfields])

        argstring = f'Topologies: {only_topology or "all"}; Particle exclusions: {exclude_particles}, {exclude_particles_pdg}; WFRs: {wfrs}; Tadpoles: {tadpoles_str}'
        # TODO: write argstring as comment to the json file
        argshash = md5(argstring.encode('utf-8')).hexdigest()
        cachefile = path.join(self.cachedir, f'insertions_{extfields}_{argshash}.json')

        loop_particles = [p for p in self.loop_particles if (p.name not in exclude_particles and abs(p.pdg_code) not in exclude_particles_pdg)]
        loop_particles = loop_particles + [p.anti() for p in loop_particles if p != p.anti()]
        if not loop_particles:
            return insertions
        caching = self.caching > 0
        if safe_isfile(cachefile) and caching:
            logger.info(f'Using insertions for {extfields} from cache ({cachefile}).')
            with open(cachefile, 'r') as f:
                insertions = json.load(f)
                insertions.pop('setup', None)
                return insertions
        if caching:
            lock = CreateLock(cachefile)
        if not only_topology:
            desc = f'determine all insertions for {extfields}-amplitude for all topologies.'
        else:
            desc = f'determine all insertions for {extfields}-amplitude for the {only_topology} toplogy.'
        logger.info(desc)
        if only_topology and only_topology in topologies:
            topologies = {only_topology: topologies[only_topology]}

        for t, topo in tqdm(topologies.items(), disable = not self.progress, desc = desc, leave = False):
            insertions[t] = []
            topo.insertexternal(*externalfields)
            # TODO: this could be optimized further. For now it works and is nor urgent (due to caching)
            num = int(factorial(len(loop_particles)+topo.nint-1)/factorial(topo.nint)/factorial(len(loop_particles)-1))
            for particles in tqdm(combinations_with_replacement(loop_particles, topo.nint), total = num, leave = False, desc = t, disable = not self.progress):
                perms = set(permutations(particles))
                for p in perms:
                    topo.insertinternal(*p)
                    if topo.tadpole is not False and topo.intfields[topo.tadpole].insertion.spin != 1:
                        # only scalar tadpoles
                        continue
                    insertion = topo.getinsertion()
                    couplings = [self.getcoupling(ins) for ins in insertion]
                    conj_couplings = [self.getcoupling([p.anti() for p in ins]) for ins in insertion]
                    if all(couplings) and all(conj_couplings):
                        insertions[t].append([[k.name for k in p], [couplings,conj_couplings]])
            logger.debug(f'Found insertions for topology {t}: {[i[0] for i in insertions[t]]}')
        logger.info(f'{sum(len(i) for i in insertions.values())} (potentially) non-zero diagrams to calculate found.')
        if caching:
            with open(cachefile, 'w') as f:
                insertions['setup'] = argstring
                json.dump(insertions, f, default = lambda x: x.name, indent = None)
                insertions.pop('setup')
            lock.release()
        return insertions

    def ColorFactor(self, externalfields: list, particles: list, topo: str):
        """ hard-coded color factors for various topologies

        Args:
            externalfields: list of field names or objects
            particles: list of particles inserted into the topology
            topo: name of topology to be considered

        Returns:
            color factor
        """

        colors = [p.color for p in particles] if particles else [0]
        extcolors = [p.color for p in externalfields]
        if topo in ['TwoPointTA','ThreePointTA','ThreePointWFRA','ThreePointWFRB']:
            colors = colors[1:]
        if topo  == 'ThreePointWFRT':
            colors = colors[2:]
        colors = tuple(sorted(colors))

        if all(c == 1 for c in extcolors):
            colorfac = {
                    (0,): 1,
                    (1,): 1,
                    (3,): 3,
                    (-3,): 3,
                    (8,): 8,
                    (1,1): 1,
                    (3,3): 3,
                    (-3,-3): 3,
                    (8,8): 8,
                    (1,1,1): 1,
                    (3,3,3): 3,
                    (-3,-3,-3): 3,
                    (8,8,8): 8
                    }.get(colors, 0)

        if extcolors in [[3,-3],[-3,3]]:
            colorfac = {
                    (0,): 1,
                    (1,): 1,
                    (3,): 3,
                    (-3,): 3,
                    (8,): 8,
                    (1,1): 1,
                    (-3,1): 1,
                    (1,3): 1,
                    (3,8): '4/3',
                    (-3,8): '4/3'
                    }.get(colors, 0)

        if colorfac == 0:
            colorfac = 1
            # TODO caclulate color factor using UFO couplings
            logger.error(f'Color structure "{colors}" not supported in diagram {topo}[{particles}]!')
        return colorfac



import cmath
from functools import lru_cache
import rundec
import logging

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

@lru_cache(maxsize=5000)
def kaellen(x,y,z):
    return (x-y-z)**2 - 4*y*z

@lru_cache(maxsize=5000)
def tminus(Q2,m12,m22):
    return -0.5*(Q2 - m12 - m22 + cmath.sqrt(kaellen(Q2,m12,m22)))

@lru_cache(maxsize=5000)
def tplus(Q2,m12,m22):
    return -0.5*(Q2 - m12 - m22 - cmath.sqrt(kaellen(Q2,m12,m22)))

@lru_cache(maxsize=5000)
def RunAlfas(Qstart, Qend, AlfasIn, MT=172.5, MB=4.18, looplevel=2):
    """Performs running of $\\alpha_s$ in the SM using RunDec.

    Args:
        Qstart: the input scale in GeV.
        Qend: the end scale in GeV.
        AlfasIn: $\\alpha_s(Q_\\text{start})$.

    Returns:
        $\\alpha_s(Q_\\text{end})$
    """

    if Qstart < MB or Qstart > MT:
        logger.error('Alfas running only implemented for Qstart > MB, Qstart < MT')
        return

    crd = rundec.CRunDec()
    if Qend < MT:
        return crd.AlphasExact(AlfasIn, Qstart, Qend, 5, looplevel)
    else:
        as5MT = crd.AlphasExact(AlfasIn, Qstart, MT, 5, looplevel)
        as6MT = crd.DecAsUpOS(as5MT, MT, 6, looplevel-1)
        return crd.AlphasExact(as6MT, MT, Qend, 6, looplevel)

    # return AlfasIn/(1 + (11 - 10/3.)/(4*cmath.pi)*AlfasIn*cmath.log((Qend/Qstart)**2))

def MZfw(MZrun, GammaZ = 2.4952):
    """Converts $M_Z$ defined with running width to $M_Z$ defined with fixed width.

    Args:
        MZrun: $M_Z$ defined with running width
        GammaZ: total decay width of the $Z$ boson

    Returns:
        $M_Z$ defined with fixed width
    """
    return MZrun - 0.5*GammaZ**2/MZrun

def MZrun(MZfw, GammaZ = 2.4952):
    """Converts $M_Z$ defined with fixed width to $M_Z$ defined with running width.

    Args:
        MZfw: $M_Z$ defined with fixed width
        GammaZ: total decay width of the $Z$ boson

    Returns:
        $M_Z$ defined with running width
    """
    return MZfw + 0.5*GammaZ**2/MZfw

def MWfw(MWrun, AlfasMW, GF):
    """Converts $M_W$ defined with running width to $M_W$ defined with fixed width.

    Args:
        MWrun: $M_W$ defined with running width
        AlfasMW: $\\alpha_s(M_W)$
        GF: the Fermi constant $G_F$

    Returns:
        $M_W$ defined with fixed width
    """
    return MWrun*(1 - (3/(4*cmath.pi)*GF*MWrun**2*(1 + 2/(3*cmath.pi)*AlfasMW))**2)

def MWrun(MWfw, AlfasMW, GF):
    """Converts $M_W$ defined with fixed width to $M_W$ defined with running width.

    Args:
        MWfw: $M_W$ defined with fixed width
        AlfasMW: $\\alpha_s(M_W)$
        GF: the Fermi constant $G_F$

    Returns:
        $M_W$ defined with running width
    """
    return MWfw*(1 + (3/(4*cmath.pi)*GF*MWfw**2*(1 + 2/(3*cmath.pi)*AlfasMW))**2)



from os import path, remove
from time import sleep
import logging
from filelock import FileLock, Timeout

__doc__ = """ # AnyLock
A module to handle file-locks and safe-loading/writing of files (mostly for caching analytic results)
"""

def safe_isfile(file, timeout = 900):
    """ Check if a file _and_ its lock exists
    Args:
        file: path to file (w/o lock)
        timeout: total time to wait for the file `file+'.lock'` to disappear
    Returns:
        result of path.isfile(file)

    If timeout has passed, the lock file is removed
    """
    lock_file = file + '.lock'
    total_time = 0
    poll_interval = 0.05
    while path.isfile(lock_file):
        # found a lock file. Wait for it to disappear
        if total_time > 0 and total_time % 10 == 0:
            logging.warning(f'found lock file {lock_file}. Waiting...')
        if total_time >= timeout:
            logging.error(f'lock file {lock_file} seems to be stale. Removing it.')
            try:
                remove(lock_file)
            except FileNotFoundError:
                pass
            return path.isfile(file)
        sleep(poll_interval)
        total_time += poll_interval
        poll_interval = min(poll_interval+0.05, 2)
    return path.isfile(file)

class CreateLock(object):
    """
    Creates a filelock.FileLock object for `lock_file` and tries to acquire a lock.

    Args:
        lock_file: path to file, including .lock file extension
        acquire_timeout: passed to timeout argument of FileLock.acquire
        timeout: total time to wait/retry acquiring the lock. Afterwards lock_file is considered stale and deleted
        open:

    Returns:
        lock object

    Example:
    ```python
    with CreateLock('path/to/file.json'):
        # code that reads/writes file.json
    # or:
    lock = CreateLock('path/file.json')
    try:
        # code in here
    finally:
        lock.release()
    ```
    """

    def __init__(self, lock_file, timeout = 900, acquire_timeout = 10):
        self.lock_file = lock_file
        self.timeout = timeout
        self.acquire_timeout = acquire_timeout

        if not lock_file.endswith('.lock'):
            self.lock_file = lock_file + '.lock'

        self.lock = FileLock(self.lock_file)
        self.lock_count = 0
        self.acquire()

    def release(self, **kwargs):
        self.lock.release(**kwargs)
        if path.isfile(self.lock_file):
            try:
                remove(self.lock_file)
            except:
                logging.warning(f'not able to clean up .lock file {self.lock_file}')

    def acquire(self):
        while not self.lock.is_locked:
            try:
                self.lock.acquire(timeout=self.acquire_timeout)
            except Timeout:
                logging.warning(f'file {self.lock_file} is locked. Waiting...')
                self.lock_count += 1
                sleep(0.1)

            if self.lock_count * (self.acquire_timeout + 0.1) > self.timeout:
                logging.error(f'stale lock file {self.lock_file}! Removing it. Consider clearing the cache to ensure consisten results!')
                try:
                    remove(self.lock_file)
                except FileNotFoundError:
                    pass
                sleep(0.1)

    def __enter__(self):
        return self.lock

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __del__(self):
        self.release(force=True)

    def __str__(self):
        return self.lock_file

class safe_open(object):
    """ drop-in replacement for `open` taking care of filelocks using `CreateLock`

    Example:
    ```python
    with safe_open('/path/to/file.json','w'):
        # code to modify file.json
    ```
    """
    def __init__(self, file, mode, CreateLock_args = {}, **kwargs):
        self.file = file
        self.mode = mode
        self.CreateLock_args = CreateLock_args
        self.kwargs = kwargs
        self.lock = CreateLock(file, **CreateLock_args)

    def __enter__(self):
        self.open = open(self.file, self.mode, **self.kwargs)
        return self.open

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.open.close()
        self.lock.release()

    def __str__(self):
        return self.file



from collections import Counter
from os import path
import sys
import gc
import importlib.util
import logging
import pylha

__doc__ = """ # AnyUtility
A module to store all kinds of useful functions used accross different modules.
"""

spintypes = {-1: 'U', 1 : 'S', 2 : 'F', 3 : 'V'}
""" Mapping between UFO integers representing spins and internal field strings """
typesspin = {'U': -1, 'S': 1, 'F': 2, 'V': 3}
""" Inverse of `spintypes` """

def fieldtypes(fieldlist, sort = False):
    """ Take a list of UFO particle objects, `anyBSM.ufo.object_library.Particle()`,
    and returns a string with the corresponding field types.
     * `sort`: whether to sort the field types before concatenating them
    Examples:
     * `fieldtypes([3,3,1])` yields "VVS"
     * `fieldtypes([3,3,2], sort=True)` yields "SVV"
    """
    types = [spintypes.get(f.spin, 'UNKOWN') for f in fieldlist]
    if sort:
        types.sort()
    return ''.join(types)

def fieldcount(fieldlist):
    """ Returns `collections.Counter()` yielding
    how often a field is contained in a list """
    return Counter([p.name for p in fieldlist])

module_dir = path.abspath(path.dirname(__file__))
""" Path to anyBSM installation """

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        choice = input(question + " (respond with 'yes' or 'no' or 'y' or 'n') " + prompt).lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]

def import_sympy():
    """ Importing sympy can by quite expensive. Furthermore, we want
    to preserve global options set in sympy across different modules.
    This function checks if sympy was already imported (via
    `sys.modules`) and returns the existing package instead.\n
    Usage: `sp = import_sympy()`"""
    if 'sympy' not in sys.modules:
        import sympy # noqa: F401
    return sys.modules['sympy']

class LHA():
    """ A simple (S)LHA parser pased on
    [pylha](https://github.com/DavidMStraub/pylha)"""
    def __init__(self,file):
        """ `file`: path to the (S)LHA file to load` """
        try:
            with open(file,'r') as f:
                lha = pylha.load(f)
        except FileNotFoundError:
            logging.error(f'File {file} not found.')
            return {}
        if 'BLOCK' not in lha:
            logging.warning(f'No (S)LHA blocks defined in {file}')
            return None
        self.blocks = {k.upper():v for k,v in lha['BLOCK'].items()}
        self.blocks['DECAY'] = lha.get('DECAY', {})
        self.file = file

    def get(self, lha_block, lha_code):
        """ Return the value from `lha_block` (string) at position
        `lha_code` (list of integers).\n `
        """
        try:
            block = self.blocks[lha_block.upper()]
        except KeyError:
            logging.error(f'LHA block {lha_block} not found in {self.file}')
            return 0
        if lha_block == 'DECAY':
            pid = str(lha_code[0])
            # return total width
            try:
                return self.blocks['DECAY'][pid]['info'][0]
            except KeyError:
                logging.warning(f'Decay for PID {pid} not found in {self.file}')
                return
        block = block['values']
        size = len(lha_code)
        for entry in block:
            if entry[:size] == lha_code:
                if len(entry[size:]) == 1:
                    return entry[size]
                return entry[size:]
        logging.error(f'LHA block {lha_block} in {self.file} does not contain the entry {lha_code}')
        return 0

def lazy_import(module_name, file_path):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if module_name in sys.modules:
        del sys.modules[module_name]
    gc.collect()
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module



from tqdm.auto import tqdm
from pandas import DataFrame
import matplotlib
import matplotlib.axes
import numpy as np
from numpy.random import default_rng
from PIL import Image
from os import path
from anyBSM import anyBSM
from anyBSM.utils import module_dir

__doc__ = """# AnyPlot
A module to set default matplotlib-styles, perform simple random- and grid-scans with anyBSM as well as to plot the results.
"""

default_labels = {
    'kappalambda' : '$\\kappa_\\lambda$',
    'total' : '$\\lambda^{\\mathrm{NLO}} \\,\\, \\mathrm{GeV}$',
    'treelevel' : '$\\lambda^{\\mathrm{LO}} \\,\\, \\mathrm{GeV}$',
    'eig0' : '$|a_{J=0}^{\\mathrm{max}}|$',
}
""" LaTeX representation of quantities that most likely appear on one
of the axes but do not happen to be defined in UFO """

logo_path = path.join(module_dir, 'logos')

anyLogo = {'resize': False, 'show': True, 'alpha': 0.5, 'zorder': 0, 'loc': 'lower right'}
""" Default values for the anyH3 logo which is automatically added to all matplotlib axes/plots.
`show` may be boolean or a list of integers marking the subplots which shall contain the logo.
"""

def logo(resize=False, logo_name = None):
    """ load logo and optionally resize it. Return logo PIL object """
    size = resize or anyLogo.get('resize', False)
    logoname = logo_name if logo_name else anyLogo.get('logo_name', 'anyH3_logo_plot.png')
    img = Image.open(path.join(logo_path, logoname))
    if size:
        img = img.resize(
            (int(size*img.size[0]),int(size*img.size[1])),
            Image.LANCZOS)
    return img

def place_logo(fig,x=0.9,y=0.69,resize=False,logo_name='anyH3_logo_plot.png',**kwargs):
    """ place logo manually to a given figure.

    It is imortant to set the figure dpi before calling this function
    and further to set the dpi size in `savefig`.
    Example:
    ```python
    fig, ax = plt.subplots()
    fig.dpi = 600
    # do your plotting here
    plotting.place_logo(fig,0.92,0.75, alpha=0.4,logo_name='anyH3_logo_small.png')
    fig.savefig('file.pdf', bbox_inches='tight',dpi=fig.dpi)
    ```
    """
    img = logo(resize=resize, logo_name=logo_name)
    if isinstance(fig, matplotlib.figure.FigureBase):
        return fig.figimage(
                img,
                x*fig.dpi*fig.get_figwidth(),
                y*fig.dpi*fig.get_figheight(),
                **kwargs)
    else:
        return img

class anyAxes(matplotlib.axes.Axes):

    """ Inherits/overwrites `matplotlib.axes.Axes` and adds the anyH3 logo to the Axes.
    The behaviour/styling of the logo can be changed using `anyBSM.anyBSM.plotting.anyLogo`.
    Alternatively use `plotting.place_logo` to place the logo manually.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if anyLogo.get('show', True) is False :
            return
        disable = anyLogo.get('show', [])
        disable = [] if disable is True else disable
        if disable and hasattr(args[1],'num1') and args[1].num1 not in disable:
            return

        img = logo()
        if isinstance(anyLogo['loc'], (list,tuple)) and hasattr(args[0],'figimage'):
            args[0].figimage(img, *anyLogo['loc'], alpha=anyLogo['alpha']) # manual placement
            return
        img_box = matplotlib.offsetbox.OffsetImage(img,alpha=anyLogo.get('alpha',0.5),dpi_cor = False)
        img_anchored = matplotlib.offsetbox.AnchoredOffsetbox(
                  anyLogo.get('loc', 'lower right'),
                  child = img_box,
                  frameon=False,
                  borderpad=0)
        img_anchored.set_zorder(anyLogo.get('zorder', 0))
        self.add_artist(img_anchored)

matplotlib.axes.Axes = anyAxes

import matplotlib.pyplot as plt # noqa: E402
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['xtick.top']   = True
matplotlib.rcParams['ytick.right'] = True
matplotlib.rcParams['xtick.minor.visible'] = True
matplotlib.rcParams['ytick.minor.visible'] = True
matplotlib.rcParams['xtick.direction'] = 'in'
matplotlib.rcParams['ytick.direction'] = 'in'
matplotlib.rcParams["xtick.minor.visible"] = True
matplotlib.rcParams["ytick.minor.visible"] = True
matplotlib.rcParams['font.size']   = 16.0
matplotlib.rcParams['font.family'] = 'serif'

def setAnyStyle(style='paper'):
    if style == 'paper':
        matplotlib.rcParams['font.family'] = 'serif'
    if style == 'beamer':
        matplotlib.rcParams['font.size']   = 18.0

class ScanParameter():
    def __init__(self, parameter, distribution = 'uniorm', *args, **kwargs):
        """ Defines a scan parameter.
          * `parameter`: anyBSM.ufo.object_library.Parameter()-object,
          i.e. any parameter defined in the UFO model
          * `distribution`: generator used in
          numpy.random.default_rng instance. Special value 'grid' is
          used to use `np.linspace` instead (in which case the kwargs
          'start' and 'stop' are required)
          * `*args`/`**kwargs` additional arguments passed to the
          generator (e.g. 'start' and 'stop' etc)
        """

        self.name = parameter.name
        self.tex = parameter.texname
        self.distribution = distribution
        self.values = []
        """ Store for the generated values """
        self.args = args
        self.kwargs = kwargs

    def generate(self, rng, n):
        """ Generate `n` random values using the generator instance
        `rng`. In case of `distribution=grid` was used, `np.linspace`
        is issued instead."""
        self.kwargs['size'] = n
        if self.distribution == 'grid':
            self.values = np.linspace(self.kwargs['start'], self.kwargs['stop'], num = n)
            return
        self.values = getattr(rng, self.distribution)(*self.args, **self.kwargs)

class Plotting():
    def __init__(self,model,lambdaSM=187.28177740658242, caching = 2, **kwargs):
        """ Perform scans using pre-defined scan parameter(ranges)
        (i.e. `ScanParameter`s).\n
         * `model`: anyBSM object
         * `lambdaSM`: normalization value for
         $\\kappa_\\lambda$-plots"""
        if not lambdaSM:
            print('calculating lambda_hhh^tree in the SM"')
            SM = anyBSM('SM',
                       progress=False,
                       quiet=True,
                       scheme_name='MS',
                       evaluation='numerical')
            SM.setparameters()
            self.lamSM = complex(0,1)*SM.getcoupling('h','h','h')['c'].nvalue
            print(f'    \\lambda_hhh^SM = {self.lamSM}')
        else:
            self.lamSM = lambdaSM

        self.model = model
        self.model.caching = caching
        model.evaluation = 'numerical'
        model.progress = False
        model.quiet = True

        self.rng = default_rng()

        self.results = {}

        setAnyStyle()

    def RandomScan(self, name = '', n = 0, parameters = [], **kwargs):
        """ Perform a random `anyBSM.lambdahhh()`-scan using a given list of scan
        `parameters` (while all other parameters are taken to be the
        default values from the UFO model). The scan runs until `n` samples have been
        calculated.
         * `name`: name of the scan
         * `n`: size of the scan sample
         * `parameters`: list of ScanParameter() objects
         * `kwargs`: passed to `anyBSM.anyBSM.lambdahhh(**kwargs)`
        The scan result is saved in a Pandas DataFrame
        `self.results[name]` and can be used/plotted with the
        `Scatter()` method.
        """
        if not name:
            name = f'scan_{len(self.results)}'
        if name in self.results:
            return self.results[name]
        for p in parameters:
            p.generate(self.rng, n)
        tmp = []
        for i in tqdm(range(n)):
            paras = {p.name: p.values[i] for p in parameters}
            self.model.setparameters(paras)
            res = self.model.lambdahhh(**kwargs)
            for k,v in res.items():
                paras.update({f'lambdahhh{k}': v})
            paras.update({'kappalambda': res['total'].real/self.lamSM, 'eig0': self.model.eigSSSS()})
            tmp.append(paras)
        self.results[name] = DataFrame(tmp)
        return self.results[name]

    def XY(self, scanpara, resultspara, n, parameters ={},lambdahhh_kwargs={}, *args, **kwargs):
        """ Perform a linear `anyBSM.lambdahhh()`-scan by varying on
        `ScanParameter()` `scanpara` and plot it against
        `resultspara`:
         * `scanpara`: a `ScanParameter()` object
         * `resultspara`: a string representing a column of the
         resulting DataFrame of the scan
         * `n`: number of points to plot
         * `parameters`: additional parameters to set; passed to
         `anyBSM.anyBSM.setparameters()`
         * `lambdahhh_kwargs`: arguments passed to anyH3.lambdahhh()
         * `*args/**kwargs`: additional arguments passed to
         `matplotlib.pyplot.plot`
        The `resultspara` value can be `lambdahhh` or 'kappalambda'. In
        addition, the individual contributions to lambdahhh
        {tree-level, genuine,WFRs,tadpoles,massCTs,VEVCT,customCT} are
        available via `lambdahhh{0,1,2,3,4,5,6}`.\n
        Example:
        ```python
        # here `SM` is the anyBSM model object
        MH = ScanParameter(SM.parameters['Mh'], distribution='grid',start = 10, stop = 500)
        plot = Plotting(SM)
        plot.XY(MH,'kappalambda', 20, parameters = {'Mu3': 173}, label='$M_{top}= 173$ GeV')
        plot.XY(MH,'kappalambda', 20, parameters = {'Mu3': 150}, label='$M_{top}= 150$ GeV')
        ```
        """
        tmp = []
        self.model.setparameters(parameters)
        scanpara.generate(self.rng, n)
        for i in range(n):
            paras = dict(parameters)
            paras = {scanpara.name: scanpara.values[i]}
            self.model.setparameters(paras)
            res = self.model.lambdahhh(**lambdahhh_kwargs)
            for k,v in res.items():
                paras.update({f'lambdahhh{k}': v})
            paras.update({'kappalambda': res['total'].real/self.lamSM, 'eig0': self.model.eigSSSS()})
            tmp.append(paras)
        df = DataFrame(tmp)
        p = plt.plot(df[scanpara.name], df[resultspara], *args, **kwargs)
        ylabel = default_labels.get(resultspara, resultspara)
        plt.xlabel(f'${scanpara.tex}$')
        plt.ylabel(ylabel)
        return p

    def Scatter(self, name, parameter_x, parameter_y, colorbar_kwargs = {}, *args, **kwargs):
        """ Plot the 2(3)-dimensional hypersphere of the result of a previous
        N-dimensional random scan `name` performed with `RandomScan(name=name)`.
         * `name`: name of the scan; stored in the DataFrame self.results[name]
         * `parameters_x`: parameter/column from the DataFrame to be
         plotted on the x-axis
         * `parameters_y`: parameter/column from the DataFrame to be
         plotted on the y-axis
         * `*args/*kwargs`: additional arguments for
         matplotlib.pyplot.scatter such as e.g. `c='kappalambda' to
         automatically  introduce a heat-map and add a colorbar with
         the label $\\kappa_\\lambda$`.\n
        Example:
        ```python
        plot = Plotting(SM)
        MT = ScanParameter(SM.parameters['Mu3'], distribution='uniform', low = 150, high = 180)
        MH = ScanParameter(SM.parameters['Mh'] , distribution='uniform', low = 100, high = 300)
        plot.RandomScan(name = 'MTMHScan', parameters = [MT,MH], n = 500)
        p = plot.Scatter('MTMHScan',MH,MT, c='kappalambda')
        ```
        """
        self.model.setparameters()
        if name not in self.results:
            print(f'cant find results for scan "{name}"')
            return
        df = self.results[name]
        if 'c' in kwargs and isinstance(kwargs['c'], str):
            clabel = default_labels.get(kwargs['c'], kwargs['c'])
            kwargs['c'] = df[kwargs['c']]
        else:
            clabel = kwargs['c']

        p = plt.scatter(df[parameter_x.name], df[parameter_y.name], *args, **kwargs)
        xlabel = parameter_x.tex.replace('text', 'rm')
        ylabel = parameter_y.tex.replace('text', 'rm')
        plt.xlabel(fr'${xlabel}$')
        plt.ylabel(fr'${ylabel}$')
        cbar = None
        if 'c' in kwargs:
            cbar = plt.colorbar(p, **colorbar_kwargs)
            cbar.set_label(clabel)
        fig = plt.gcf()
        return [fig, cbar]



#!/usr/bin/env python3
import unittest
from anyBSM import anyBSM
import numpy as np
import numpy.testing as test
""" This is a file with tests for 1L corrections to la112 in the THDM type II, The numbers for the numerical check were obtained in the notebook test_la112.nb analytic expressions were not included because they are too large """

class numericTest(unittest.TestCase):

    def setUp(self):
        self.THDM = anyBSM(
            anyBSM.built_in_models['THDMII'], # load the model
            scheme_name = "OSFJ", # load the OS scheme
            progress = False, # dont show any progress bars
            caching = 0, # disable the cache
            quiet=True
        )
        self.THDM.clear_cache()
        # set parameters for comparison (the check will only work for those values because it is a numerical check)
        self.THDM.setparameters({'alphaH': np.arctan(2)-np.arcsin(0.9), 'TanBeta': 2, 'Mh1': 125.1, 'Mh2': 300., 'MAh2': 300., 'MHm2': 550., 'M': 200})

    """ General 1L corrections to the 112 coupling in the THDM  (only light fermions are excluded) """
    def test_genuine_la112(self):
        """ define the particles and their moemnta """
        p = self.THDM.all_particles
        hl = p['h1']
        hh = p['h2']
        p1 = 125.1**2
        p2 = 125.1**2
        p3 = 300.**2
        """ THREE POINT B """
        hhHTPBanyH3 = self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']], only_topologies=['ThreePointB'])['ThreePointB']
        hhHTPBmath = -257.68536481277926+6.2071520452980025j
        test.assert_allclose(hhHTPBanyH3, hhHTPBmath)
        """ THREE POINT C """
        hhHTPCanyH3 = self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']], only_topologies=['ThreePointC'])['ThreePointC']
        hhHTPCmath = -75.00919028357599-8.506510393334167j
        test.assert_allclose(hhHTPCanyH3, hhHTPCmath)
        """ GENUINE (sum of the previous two) """
        hhHgenanyH3 = hhHTPBanyH3 + hhHTPCanyH3
        hhHgenmath = -332.69455509635526-2.299358348036165j
        test.assert_allclose(hhHgenanyH3, hhHgenmath)

    def test_ta_la112(self):
        """ define the particles and their momenta """
        p = self.THDM.all_particles
        hl = p['h1']
        hh = p['h2']
        p1 = 125.1**2
        p2 = 125.1**2
        p3 = 300.**2
        """ THREE POINT TA """
        hhHTPTAanyH3 = self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']], only_topologies=['ThreePointTA'])['ThreePointTA']
        hhHTPTAmath = 28.511383412114707
        test.assert_allclose(hhHTPTAanyH3, hhHTPTAmath)

    def test_wfra_la112(self):
        """ define the particles and their momenta """
        p = self.THDM.all_particles
        hl = p['h1']
        hh = p['h2']
        p1 = 125.1**2
        p2 = 125.1**2
        p3 = 300.**2
        """ Three point WFRA """
        hhHTPWFRanyH3 = self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']],only_topologies=['ThreePointWFRA'])['ThreePointWFRA'] + self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']],only_topologies=['ThreePointWFRB'])['ThreePointWFRB'] + self.THDM.process(hl,hl,hh,momenta=[p1,p2,p3],exclude_particles=[p for p in ['u2','u1','d3','d2','d1','u2bar','u1bar','d3bar','d2bar','d1bar','e1','e2','e3','e1bar','e2bar','e3bar']],only_topologies=['ThreePointWFRT'])['ThreePointWFRT']
        hhHTPWFRmath = 38.706861622118595-0.07839740766875303j
        test.assert_allclose(hhHTPWFRanyH3, hhHTPWFRmath)

if __name__ == '__main__':
    unittest.main()



__doc__ = """
# (Modified) UFO objects.

This is the standard UFO library. However, a few improvements have
been performed:

 * The `global all_<ufo object>` variables are dictionaries instead of lists
   This has no effect on the models initialization (particles.py etc)
   but only on how the objects are stored internally (e.g. the particle
   with name `h1` is directly accessible via `all_particles['h1']`).
 * New method `UFOBaseClass.dump()`: returns string representation to
 be interpreted by python
 * UFOBaseClass now warns if duplicate objects (same .name) are initialized
 * `nvalue` attributes have been introduced for Parameter() objects.
 They contain the numerical value obtained with the current set of
 `values` of all  external parameters.
 * `nmass` similar to `nvalue` but for Particle() objects.
 * `Particle.anti()` has been changed to avoid the duplicate creation
 of Particle() objects but rather returns the corresponding anti
 particle from `all_particles`.
 * `function_library` was completely rewritten to avoid `__exec__`ing strings of functions.
 This yields a significat performance boost for large models of up to 3 orders of magnitued in runtime.

Everything else is identical to the original UFO standard.
"""
